import pathlib
from collections import defaultdict
from typing import Any, TypedDict, cast

import pydantic_core
import tqdm
from eth_typing import ABIEvent, ChecksumAddress, HexAddress
from eth_utils.abi import event_abi_to_log_topic
from hexbytes import HexBytes
from web3 import Web3
from web3.contract.base_contract import BaseContractEvent
from web3.types import EventData, LogReceipt
from web3.utils import get_abi_element

from degenbot import connection_manager, get_checksum_address
from degenbot.functions import fetch_logs_retrying
from degenbot.logging import logger
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import KeyedDefaultDict
from degenbot.uniswap.abi import UNISWAP_V3_POOL_ABI
from degenbot.uniswap.v3_types import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolLiquidityMappingUpdate,
)


class LiquidityMap(TypedDict):
    tick_bitmap: dict[int, UniswapV3BitmapAtWord]
    tick_data: dict[int, UniswapV3LiquidityAtTick]


class UniswapV3LiquiditySnapshot:
    """
    Retrieve and maintain liquidity positions for Uniswap V3 pools.
    """

    def __init__(
        self,
        path: pathlib.Path | str,
        chain_id: ChainId | None = None,
    ) -> None:
        def _get_pool_map(pool_address: HexAddress | bytes) -> LiquidityMap:
            """
            Get the liquidity map for the pool.
            """

            pool_address = get_checksum_address(pool_address)

            if self._file_snapshot is not None and pool_address in self._file_snapshot:
                return LiquidityMap(
                    tick_bitmap={
                        int(k): UniswapV3BitmapAtWord(**v)
                        for k, v in self._file_snapshot[pool_address]["tick_bitmap"].items()
                    },
                    tick_data={
                        int(k): UniswapV3LiquidityAtTick(**v)
                        for k, v in self._file_snapshot[pool_address]["tick_data"].items()
                    },
                )
            if self._dir_path is not None and (
                (pool_map_file := self._dir_path / f"{pool_address}.json").exists()
            ):
                pool_liquidity_snapshot = pydantic_core.from_json(pool_map_file.read_bytes())
                return LiquidityMap(
                    tick_bitmap={
                        int(k): UniswapV3BitmapAtWord(**v)
                        for k, v in pool_liquidity_snapshot["tick_bitmap"].items()
                    },
                    tick_data={
                        int(k): UniswapV3LiquidityAtTick(**v)
                        for k, v in pool_liquidity_snapshot["tick_data"].items()
                    },
                )

            return LiquidityMap(
                tick_bitmap={},
                tick_data={},
            )

        path = pathlib.Path(path).expanduser()
        assert path.exists()

        self.newest_block: BlockNumber
        self._dir_path: pathlib.Path | None = None
        self._file_snapshot: dict[str, Any] | None = None

        if path.is_file():
            self._file_snapshot = pydantic_core.from_json(path.read_bytes())
            self.newest_block = self._file_snapshot.pop("snapshot_block")
        if path.is_dir():
            self._dir_path = path
            metadata_path = path / "_metadata.json"
            assert metadata_path.exists()
            self.newest_block = pydantic_core.from_json(metadata_path.read_bytes())["block"]

        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id

        logger.info(
            f"Loaded Uniswap V3 LP snapshot: {len(self._file_snapshot) if self._file_snapshot else len(tuple(path.glob('0x*.json')))} pools @ block {self.newest_block}"  # noqa:E501
        )

        self._liquidity_events: dict[ChecksumAddress, list[UniswapV3LiquidityEvent]] = defaultdict(
            list
        )
        self._liquidity_snapshot: dict[ChecksumAddress, LiquidityMap] = KeyedDefaultDict(
            _get_pool_map
        )

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def pools(self) -> set[ChecksumAddress]:
        if self._file_snapshot is not None:
            return {get_checksum_address(pool) for pool in self._file_snapshot} | {
                get_checksum_address(pool) for pool in self._liquidity_events
            }
        if self._dir_path is not None:
            return {
                get_checksum_address(pool_filename.stem)
                for pool_filename in self._dir_path.glob("0x*.json")
            }
        raise RuntimeError  # pragma: no cover

    def fetch_new_events(
        self,
        to_block: BlockNumber,
        blocks_per_request: int | None = None,
    ) -> None:
        """
        Fetch liquidity events from the block following the last-known event to the target block
        using `eth_getLogs`. Blocks per request will be capped at `blocks_per_request`.
        """

        def _process_liquidity_event_log(
            event: BaseContractEvent, log: LogReceipt
        ) -> tuple[ChecksumAddress, UniswapV3LiquidityEvent]:
            """
            Decode an event log and convert to an address and a `UniswapV3LiquidityEvent` for
            processing with `UniswapV3Pool.update_liquidity_map`.
            """

            # ref: https://github.com/Uniswap/v3-core/blob/main/contracts/interfaces/pool/IUniswapV3PoolEvents.sol
            #
            # event Mint(
            #     address sender,
            #     address indexed owner,
            #     int24 indexed tickLower,
            #     int24 indexed tickUpper,
            #     uint128 amount,
            #     uint256 amount0,
            #     uint256 amount1
            # );
            # event Burn(
            #     address indexed owner,
            #     int24 indexed tickLower,
            #     int24 indexed tickUpper,
            #     uint128 amount,
            #     uint256 amount0,
            #     uint256 amount1
            # );

            decoded_event: EventData = event.process_log(log)
            pool_address = get_checksum_address(decoded_event["address"])
            tx_index = decoded_event["transactionIndex"]
            log_index = decoded_event["logIndex"]
            liquidity_block = decoded_event["blockNumber"]
            liquidity = decoded_event["args"]["amount"]
            if decoded_event["event"] == "Burn":
                liquidity = -liquidity  # liquidity removal
            tick_lower = decoded_event["args"]["tickLower"]
            tick_upper = decoded_event["args"]["tickUpper"]

            return (
                pool_address,
                UniswapV3LiquidityEvent(
                    block_number=liquidity_block,
                    tx_index=tx_index,
                    log_index=log_index,
                    liquidity=liquidity,
                    tick_lower=tick_lower,
                    tick_upper=tick_upper,
                ),
            )

        logger.info(f"Updating Uniswap V3 snapshot from block {self.newest_block} to {to_block}")

        v3pool = Web3().eth.contract(abi=UNISWAP_V3_POOL_ABI)
        for event in (
            v3pool.events.Mint,
            v3pool.events.Burn,
        ):
            event_instance = event()
            event_abi = cast(
                "ABIEvent",
                get_abi_element(
                    abi=v3pool.abi,
                    abi_element_identifier=event.event_name,
                ),
            )

            event_logs = fetch_logs_retrying(
                w3=connection_manager.get_web3(self.chain_id),
                start_block=self.newest_block + 1,
                end_block=to_block,
                max_blocks_per_request=blocks_per_request,
                topic_signature=[HexBytes(event_abi_to_log_topic(event_abi))],
            )

            for event_log in tqdm.tqdm(
                event_logs,
                desc=f"Processing {event.event_name} events",
                unit="event",
                bar_format="{desc}: {percentage:3.1f}% |{bar}| {n_fmt}/{total_fmt}",
            ):
                pool_address, liquidity_event = _process_liquidity_event_log(
                    event_instance, event_log
                )
                self._liquidity_events[pool_address].append(liquidity_event)

        self.newest_block = to_block

    def pending_updates(
        self,
        pool_address: HexAddress,
    ) -> tuple[UniswapV3PoolLiquidityMappingUpdate, ...]:
        """
        Consume pending liquidity updates for the pool, sorted chronologically.
        """

        pool_key = get_checksum_address(pool_address)
        pending_events = tuple(self._liquidity_events[pool_key])
        self._liquidity_events[pool_key] = []

        # Mint/Burn events must be applied in chronological order to ensure effective liquidity
        # invariant checks. Sort events by block, then by log order within the block.
        sorted_events = sorted(
            pending_events,
            key=lambda event: (event.block_number, event.tx_index, event.log_index),
        )

        return tuple(
            UniswapV3PoolLiquidityMappingUpdate(
                block_number=event.block_number,
                liquidity=event.liquidity,
                tick_lower=event.tick_lower,
                tick_upper=event.tick_upper,
            )
            for event in sorted_events
        )

    def tick_bitmap(self, pool_address: str | bytes) -> dict[int, UniswapV3BitmapAtWord]:
        """
        Consume the tick initialization bitmaps for the pool.
        """

        pool_address = get_checksum_address(pool_address)
        tick_bitmap = self._liquidity_snapshot[pool_address]["tick_bitmap"]
        self._liquidity_snapshot[pool_address]["tick_bitmap"] = {}
        return tick_bitmap

    def tick_data(self, pool_address: str | bytes) -> dict[int, UniswapV3LiquidityAtTick]:
        """
        Consume the liquidity mapping for the pool.
        """

        pool_key = get_checksum_address(pool_address)
        tick_data = self._liquidity_snapshot[pool_key]["tick_data"]
        self._liquidity_snapshot[pool_key]["tick_data"] = {}
        return tick_data

    def update(
        self,
        pool: HexAddress,
        tick_data: dict[int, UniswapV3LiquidityAtTick],
        tick_bitmap: dict[int, UniswapV3BitmapAtWord],
    ) -> None:
        """
        Update the liquidity mapping for the pool.
        """

        pool_key = get_checksum_address(pool)
        self._liquidity_snapshot[pool_key]["tick_bitmap"].update(tick_bitmap)
        self._liquidity_snapshot[pool_key]["tick_data"].update(tick_data)



from degenbot.balancer.libraries.constants import ONE
from degenbot.balancer.libraries.fixed_point import (
    complement,
    div_down,
    div_up,
    mul_down,
    mul_up,
    pow_down,
    pow_up,
)
from degenbot.exceptions.evm import EVMRevertError

_MIN_WEIGHT = int(0.01 * 10**18)
# Having a minimum normalized weight imposes a limit on the maximum number of tokens;
# i.e., the largest possible pool is one where all tokens have exactly the minimum weight.
_MAX_WEIGHTED_TOKENS = 100

# Pool limits that arise from limitations in the fixed point power function (and the imposed 1:100
# maximum weight ratio).

# Swap limits: amounts swapped may not be larger than this percentage of total balance.
_MAX_IN_RATIO = int(0.3 * 10**18)
_MAX_OUT_RATIO = int(0.3 * 10**18)


# Invariant growth limit: non-proportional joins cannot cause the invariant to increase by more than
# this ratio.
_MAX_INVARIANT_RATIO = 3 * 10**18
# Invariant shrink limit: non-proportional exits cannot cause the invariant to decrease by less than
# this ratio.
_MIN_INVARIANT_RATIO = int(0.7 * 10**18)


def calculate_invariant(normalized_weights: list[int], balances: list[int]) -> int:
    invariant = ONE

    for i in range(len(normalized_weights)):
        invariant = mul_down(invariant, pow_down(balances[i], normalized_weights[i]))

    if invariant <= 0:
        raise EVMRevertError(error="ZERO_INVARIANT")

    return invariant


def _calc_out_given_in(
    balance_in: int,
    weight_in: int,
    balance_out: int,
    weight_out: int,
    amount_in: int,
) -> int:
    """
    Computes how many tokens can be taken out of a pool if `amountIn` are sent, given the
    current balances and weights.
    """

    # ********************************************************************************************
    # outGivenIn                                                                                //
    # aO = amountOut                                                                            //
    # bO = balanceOut                                                                           //
    # bI = balanceIn              /      /            bI             \    (wI / wO) \           //
    # aI = amountIn    aO = bO * |  1 - | --------------------------  | ^            |          //
    # wI = weightIn               \      \       ( bI + aI )         /              /           //
    # wO = weightOut                                                                            //
    # *******************************************************************************************/

    # Amount out, so we round down overall.

    # The multiplication rounds down, and the subtrahend (power) rounds up (so the base rounds up
    # too). Because bI / (bI + aI) <= 1, the exponent rounds down.

    # // Cannot exceed maximum in ratio
    if amount_in > mul_down(
        balance_in,
        _MAX_IN_RATIO,
    ):
        raise EVMRevertError(error="MAX_IN_RATIO")

    denominator = balance_in + amount_in
    base = div_up(balance_in, denominator)
    exponent = div_down(weight_in, weight_out)
    power = pow_up(base, exponent)
    return int(mul_down(balance_out, complement(power)))


def _calc_in_given_out(
    balance_in: int,
    weight_in: int,
    balance_out: int,
    weight_out: int,
    amount_out: int,
) -> int:
    """
    Computes how many tokens must be sent to a pool in order to take `amountOut`, given the
    current balances and weights.
    """

    # ********************************************************************************************
    # inGivenOut                                                                                //
    # aO = amountOut                                                                            //
    # bO = balanceOut                                                                           //
    # bI = balanceIn              /  /            bO             \    (wO / wI)      \          //
    # aI = amountIn    aI = bI * |  | --------------------------  | ^            - 1  |         //
    # wI = weightIn               \  \       ( bO - aO )         /                   /          //
    # wO = weightOut                                                                            //
    # *******************************************************************************************/

    # Amount in, so we round up overall.

    # The multiplication rounds up, and the power rounds up (so the base rounds up too).
    # Because b0 / (b0 - a0) >= 1, the exponent rounds up.

    _balance_in = balance_in
    _weight_in = weight_in
    _balance_out = balance_out
    _weight_out = weight_out
    _amount_out = amount_out

    # Cannot exceed maximum out ratio
    if not (_amount_out <= mul_down(_balance_out, _MAX_OUT_RATIO)):
        raise EVMRevertError(error="MAX_OUT_RATIO")

    _base = div_up(_balance_out, _balance_out - _amount_out)
    _exponent = div_up(_weight_out, _weight_in)
    _power = pow_up(_base, _exponent)
    # Because the base is larger than one (and the power rounds up), the power should always be
    # larger than one, so the following subtraction should never revert.
    _ratio = _power - ONE
    return mul_up(_balance_in, _ratio)


def _subtract_swap_fee_amount(amount: int, fee_percentage: int) -> int:
    """
    Subtracts swap fee amount from `amount`, returning a lower value.
    """

    # This returns amount - fee amount, so we round up (favoring a higher fee amount).
    fee_amount = mul_up(amount, fee_percentage)
    return amount - fee_amount



from degenbot.swapbased.pools import SwapbasedV2Pool
from degenbot.uniswap.managers import UniswapV2PoolManager


class SwapbasedV2PoolManager(UniswapV2PoolManager, pool_factory=SwapbasedV2Pool):
    type Pool = SwapbasedV2Pool



import contextlib
from threading import Lock
from typing import TYPE_CHECKING, Any, Self

from eth_typing import ChecksumAddress

from degenbot import get_checksum_address
from degenbot.connection import connection_manager
from degenbot.exceptions.liquidity_pool import LiquidityPoolError
from degenbot.exceptions.manager import (
    ManagerAlreadyInitialized,
    PoolCreationFailed,
    PoolNotAssociated,
)
from degenbot.logging import logger
from degenbot.registry import pool_registry
from degenbot.types.abstract import AbstractPoolManager
from degenbot.types.aliases import ChainId
from degenbot.uniswap.deployments import (
    FACTORY_DEPLOYMENTS,
    UniswapV2ExchangeDeployment,
    UniswapV3ExchangeDeployment,
)
from degenbot.uniswap.v2_functions import generate_v2_pool_address
from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v3_functions import generate_v3_pool_address
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool
from degenbot.uniswap.v3_snapshot import UniswapV3LiquiditySnapshot


class AbstractUniswapV2PoolManager[Pool: UniswapV2Pool](AbstractPoolManager[Pool]):
    def __init__(
        self,
        factory_address: str,
        *,
        chain_id: ChainId | None = None,
        deployer_address: ChecksumAddress | str | None = None,
        pool_init_hash: str | None = None,
    ) -> None:
        factory_address = get_checksum_address(factory_address)

        if chain_id is None:
            chain_id = connection_manager.default_chain_id

        if (chain_id, factory_address) in self.instances:
            raise ManagerAlreadyInitialized(
                message="A manager has already been initialized for this address. Access it using the get_instance() class method"  # noqa:E501
            )
        self.instances[(chain_id, factory_address)] = self

        try:
            factory_deployment = FACTORY_DEPLOYMENTS[chain_id][factory_address]
            deployer_address = (
                factory_deployment.deployer
                if factory_deployment.deployer is not None
                else factory_address
            )
            pool_init_hash = factory_deployment.pool_init_hash
        except KeyError:
            if pool_init_hash is None:  # pragma: no branch
                logger.info("Pool init hash is unknown. Using Uniswap V3 mainnet default.")
                pool_init_hash = UniswapV2Pool.UNISWAP_V2_MAINNET_POOL_INIT_HASH
            deployer_address = (
                get_checksum_address(deployer_address)
                if deployer_address is not None
                else factory_address
            )

        self._lock = Lock()
        self._chain_id = chain_id
        self._deployer_address = deployer_address
        self._factory_address = factory_address
        self._pool_init_hash = pool_init_hash
        self._tracked_pools = {}
        self._untracked_pools: set[ChecksumAddress] = set()

    def get_pool(
        self,
        pool_address: ChecksumAddress | str,
        *,
        silent: bool = False,
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> Pool:
        """
        Get a pool from its address. If the pool is already tracked or found in the global registry,
        that instance will be returned. Otherwise, a new one will be built.
        """

        pool_address = get_checksum_address(pool_address)

        with contextlib.suppress(KeyError):
            return self._tracked_pools[pool_address]

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(pool_address)

        # Check if the pool registry already has this pool
        if (
            pool_from_registry := pool_registry.get(
                pool_address=pool_address,
                chain_id=self.chain_id,
            )
        ) is not None:
            if TYPE_CHECKING:
                assert isinstance(pool_from_registry, self.pool_factory)
            if pool_from_registry.factory == self._factory_address:
                self._add_tracked_pool(pool_from_registry)
                return pool_from_registry
            self._untracked_pools.add(pool_address)
            raise PoolNotAssociated(pool_address)

        try:
            new_pool = self.pool_factory(
                address=pool_address,
                silent=silent,
                **(pool_class_kwargs or {}),
            )
        except LiquidityPoolError as exc:
            raise PoolCreationFailed(
                message=f"Could not build V2 pool {pool_address}: {exc}"
            ) from exc
        else:
            self._add_tracked_pool(new_pool)
            return new_pool


class UniswapV2PoolManager(AbstractUniswapV2PoolManager[UniswapV2Pool], pool_factory=UniswapV2Pool):
    """
    A class that generates and tracks concrete instances of a Uniswap V2 liquidity pool helper or
    one of its child classes.
    """

    @classmethod
    def from_exchange(
        cls,
        exchange: UniswapV2ExchangeDeployment,
    ) -> Self:
        return cls(
            factory_address=exchange.factory.address,
            deployer_address=exchange.factory.deployer,
            pool_init_hash=exchange.factory.pool_init_hash,
        )

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(factory={self._factory_address})"

    @property
    def pool_init_hash(self) -> str:
        return self._pool_init_hash

    def get_pool_from_tokens(
        self,
        token_addresses: tuple[str, str],
        *,
        silent: bool = False,
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> UniswapV2Pool:
        """
        Get a pool by its token addresses
        """

        pool_address = generate_v2_pool_address(
            deployer_address=self._deployer_address,
            token_addresses=token_addresses,
            init_hash=self._pool_init_hash,
        )

        return self.get_pool(
            pool_address=pool_address,
            silent=silent,
            pool_class_kwargs=pool_class_kwargs,
        )


class AbstractUniswapV3PoolManager[Pool: UniswapV3Pool](AbstractPoolManager[Pool]):
    _chain_id: ChainId
    _deployer_address: ChecksumAddress
    _factory_address: ChecksumAddress
    _tracked_pools: dict[ChecksumAddress, Pool]
    _untracked_pools: set[ChecksumAddress]
    _lock: Lock
    _pool_init_hash: str
    _snapshot: UniswapV3LiquiditySnapshot | None

    def __init__(
        self,
        factory_address: ChecksumAddress | str,
        deployer_address: ChecksumAddress | str | None = None,
        chain_id: ChainId | None = None,
        pool_init_hash: str | None = None,
        snapshot: UniswapV3LiquiditySnapshot | None = None,
    ) -> None:
        if chain_id is None:
            chain_id = connection_manager.default_chain_id

        factory_address = get_checksum_address(factory_address)

        if (chain_id, factory_address) in self.instances:
            raise ManagerAlreadyInitialized(
                message="A manager has already been initialized for this address. Access it using the get_instance() class method"  # noqa:E501
            )
        self.instances[(chain_id, factory_address)] = self

        try:
            factory_deployment = FACTORY_DEPLOYMENTS[chain_id][factory_address]
            deployer_address = (
                factory_deployment.deployer
                if factory_deployment.deployer is not None
                else factory_address
            )
            pool_init_hash = factory_deployment.pool_init_hash
        except KeyError:
            if pool_init_hash is None:  # pragma: no branch
                logger.info("Pool init hash is unknown. Using Uniswap V3 mainnet default.")
                pool_init_hash = UniswapV3Pool.UNISWAP_V3_MAINNET_POOL_INIT_HASH
            deployer_address = (
                get_checksum_address(deployer_address)
                if deployer_address is not None
                else factory_address
            )

        self._lock = Lock()
        self._chain_id = chain_id
        self._factory_address = factory_address
        self._deployer_address = deployer_address
        self._pool_init_hash = pool_init_hash
        self._snapshot = snapshot
        self._tracked_pools = {}
        self._untracked_pools = set()

    def _apply_pending_liquidity_updates(self, pool: UniswapV3Pool) -> None:
        """
        Apply all pending updates from the snapshot.
        """

        if self._snapshot:
            for liquidity_update in self._snapshot.pending_updates(pool.address):
                pool.update_liquidity_map(liquidity_update)

    def _build_pool(
        self,
        pool_address: ChecksumAddress,
        *,
        silent: bool,
        pool_class_kwargs: dict[str, Any] | None,
    ) -> Pool:
        if pool_class_kwargs is None:
            pool_class_kwargs = {}

        if self._snapshot is not None:
            pool = self.pool_factory(
                address=pool_address,
                tick_bitmap=self._snapshot.tick_bitmap(pool_address),
                tick_data=self._snapshot.tick_data(pool_address),
                silent=silent,
                **pool_class_kwargs,
            )
        else:
            logger.info("Initializing pool without liquidity snapshot")
            pool = self.pool_factory(
                address=pool_address,
                silent=silent,
                **pool_class_kwargs,
            )
        return pool

    def get_pool(
        self,
        pool_address: ChecksumAddress | str,
        *,
        silent: bool = False,
        # keyword arguments passed to the pool class constructor
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> Pool:
        """
        Get a pool from its address. If the pool is already tracked or found in the global registry,
        that instance will be returned. Otherwise, a new one will be built.
        """

        pool_address = get_checksum_address(pool_address)

        with contextlib.suppress(KeyError):
            return self._tracked_pools[pool_address]

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(pool_address)

        # Check if the pool registry already has this pool
        pool_from_registry = pool_registry.get(
            pool_address=pool_address,
            chain_id=self.chain_id,
        )
        if pool_from_registry is not None:
            if TYPE_CHECKING:
                assert isinstance(pool_from_registry, self.pool_factory)
            if pool_from_registry.factory == self._factory_address:
                self._add_tracked_pool(pool_from_registry)
                return pool_from_registry
            self._untracked_pools.add(pool_address)
            raise PoolNotAssociated(pool_address)

        try:
            new_pool = self._build_pool(
                pool_address=pool_address,
                silent=silent,
                pool_class_kwargs=pool_class_kwargs,
            )
        except LiquidityPoolError as exc:  # pragma: no cover
            raise PoolCreationFailed(
                message=f"Could not build V3 pool {pool_address}: {exc}"
            ) from exc
        else:
            self._apply_pending_liquidity_updates(new_pool)
            self._add_tracked_pool(new_pool)
            return new_pool

    def unload_snapshot(self) -> None:
        self._snapshot = None


class UniswapV3PoolManager(AbstractUniswapV3PoolManager[UniswapV3Pool], pool_factory=UniswapV3Pool):
    """
    A class that generates and tracks concrete instances of a Uniswap V3 liquidity pool helper or
    one of its child classes.
    """

    @classmethod
    def from_exchange(
        cls,
        exchange: UniswapV3ExchangeDeployment,
        snapshot: UniswapV3LiquiditySnapshot | None = None,
    ) -> Self:
        return cls(
            factory_address=exchange.factory.address,
            deployer_address=exchange.factory.deployer,
            chain_id=exchange.chain_id,
            snapshot=snapshot,
        )

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(factory={self._factory_address})"

    def _build_pool(
        self,
        pool_address: ChecksumAddress,
        *,
        silent: bool,
        pool_class_kwargs: dict[str, Any] | None,
    ) -> UniswapV3Pool:
        if pool_class_kwargs is None:
            pool_class_kwargs = {}

        if self._snapshot is not None:
            return UniswapV3Pool(
                address=pool_address,
                tick_bitmap=self._snapshot.tick_bitmap(pool_address),
                tick_data=self._snapshot.tick_data(pool_address),
                silent=silent,
                **pool_class_kwargs,
            )

        logger.info("Initializing pool without liquidity snapshot")
        return UniswapV3Pool(
            address=pool_address,
            silent=silent,
            **pool_class_kwargs,
        )

    def get_pool(
        self,
        pool_address: ChecksumAddress | str,
        *,
        silent: bool = False,
        # keyword arguments passed to the pool class constructor
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> UniswapV3Pool:
        """
        Get a pool from its address. If the pool is already tracked or found in the global registry,
        that instance will be returned. Otherwise, a new one will be built.
        """

        pool_address = get_checksum_address(pool_address)

        with contextlib.suppress(KeyError):
            return self._tracked_pools[pool_address]

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(pool_address)

        # Check if the pool registry already has this pool
        pool_from_registry = pool_registry.get(
            pool_address=pool_address,
            chain_id=self.chain_id,
        )
        if pool_from_registry is not None:
            if TYPE_CHECKING:
                assert isinstance(pool_from_registry, UniswapV3Pool)
            if pool_from_registry.factory == self._factory_address:
                self._add_tracked_pool(pool_from_registry)
                return pool_from_registry
            self._untracked_pools.add(pool_address)
            raise PoolNotAssociated(pool_address)

        try:
            new_pool = self._build_pool(
                pool_address=pool_address,
                silent=silent,
                pool_class_kwargs=pool_class_kwargs,
            )
        except LiquidityPoolError as exc:  # pragma: no cover
            raise PoolCreationFailed(
                message=f"Could not build V3 pool {pool_address}: {exc}"
            ) from exc
        else:
            self._apply_pending_liquidity_updates(new_pool)
            self._add_tracked_pool(new_pool)
            return new_pool

    def get_pool_from_tokens_and_fee(
        self,
        token_addresses: tuple[
            ChecksumAddress | str,
            ChecksumAddress | str,
        ],
        pool_fee: int,
        *,
        silent: bool = False,
        # keyword arguments passed to the pool class constructor
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> UniswapV3Pool:
        """
        Get a pool by its token addresses
        """

        pool_address = generate_v3_pool_address(
            token_addresses=sorted(token_addresses),
            fee=pool_fee,
            deployer_address=self._deployer_address,
            init_hash=self._pool_init_hash,
        )

        return self.get_pool(
            pool_address=pool_address,
            silent=silent,
            pool_class_kwargs=pool_class_kwargs,
        )



from importlib.metadata import version

# isort: split

from .checksum_cache import get_checksum_address
from .config import settings
from .connection import (
    async_connection_manager,
    connection_manager,
    get_async_web3,
    get_web3,
    set_async_web3,
    set_web3,
)

# isort: split

from .aerodrome import (
    AerodromeV2Pool,
    AerodromeV2PoolManager,
    AerodromeV2PoolState,
    AerodromeV3Pool,
    AerodromeV3PoolManager,
    AerodromeV3PoolState,
)
from .anvil_fork import AnvilFork
from .arbitrage import ArbitrageCalculationResult, UniswapCurveCycle, UniswapLpCycle
from .builder_endpoint import BuilderEndpoint
from .camelot import CamelotLiquidityPool
from .chainlink import ChainlinkPriceContract
from .curve import (
    CurveStableswapPool,
    CurveStableswapPoolSimulationResult,
    CurveStableswapPoolState,
    CurveStableSwapPoolStateUpdated,
)
from .erc20 import Erc20Token, Erc20TokenManager, EtherPlaceholder
from .logging import logger
from .pancakeswap import PancakeV2Pool, PancakeV2PoolManager, PancakeV3Pool, PancakeV3PoolManager
from .registry import pool_registry, token_registry
from .sushiswap import (
    SushiswapV2Pool,
    SushiswapV2PoolManager,
    SushiswapV3Pool,
    SushiswapV3PoolManager,
)
from .swapbased import SwapbasedV2Pool, SwapbasedV2PoolManager
from .transaction import UniswapTransaction
from .uniswap import (
    UniswapV2Pool,
    UniswapV2PoolExternalUpdate,
    UniswapV2PoolManager,
    UniswapV2PoolSimulationResult,
    UniswapV2PoolState,
    UniswapV3LiquiditySnapshot,
    UniswapV3Pool,
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolManager,
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
    UniswapV4LiquiditySnapshot,
    UniswapV4Pool,
    UniswapV4PoolExternalUpdate,
    UniswapV4PoolState,
)

__all__ = (
    "AerodromeV2Pool",
    "AerodromeV2PoolManager",
    "AerodromeV2PoolState",
    "AerodromeV3Pool",
    "AerodromeV3PoolManager",
    "AerodromeV3PoolState",
    "AnvilFork",
    "ArbitrageCalculationResult",
    "BuilderEndpoint",
    "CamelotLiquidityPool",
    "ChainlinkPriceContract",
    "CurveStableSwapPoolStateUpdated",
    "CurveStableswapPool",
    "CurveStableswapPoolSimulationResult",
    "CurveStableswapPoolState",
    "Erc20Token",
    "Erc20TokenManager",
    "EtherPlaceholder",
    "PancakeV2Pool",
    "PancakeV2PoolManager",
    "PancakeV3Pool",
    "PancakeV3PoolManager",
    "SushiswapV2Pool",
    "SushiswapV2PoolManager",
    "SushiswapV3Pool",
    "SushiswapV3PoolManager",
    "SwapbasedV2Pool",
    "SwapbasedV2PoolManager",
    "UniswapCurveCycle",
    "UniswapLpCycle",
    "UniswapTransaction",
    "UniswapV2Pool",
    "UniswapV2PoolExternalUpdate",
    "UniswapV2PoolManager",
    "UniswapV2PoolSimulationResult",
    "UniswapV2PoolState",
    "UniswapV3LiquiditySnapshot",
    "UniswapV3Pool",
    "UniswapV3PoolExternalUpdate",
    "UniswapV3PoolManager",
    "UniswapV3PoolSimulationResult",
    "UniswapV3PoolState",
    "UniswapV4LiquiditySnapshot",
    "UniswapV4Pool",
    "UniswapV4PoolExternalUpdate",
    "UniswapV4PoolState",
    "aerodrome",
    "arbitrage",
    "async_connection_manager",
    "balancer",
    "camelot",
    "cli",
    "connection_manager",
    "constants",
    "curve",
    "erc20",
    "exceptions",
    "functions",
    "get_async_web3",
    "get_checksum_address",
    "get_web3",
    "logger",
    "managers",
    "pancakeswap",
    "pool_registry",
    "registry",
    "set_async_web3",
    "set_web3",
    "settings",
    "solidly",
    "sushiswap",
    "token_registry",
    "transaction",
    "types",
    "uniswap",
    "utils",
    "validation",
)
__version__ = version(__package__)



from eth_typing import ChecksumAddress

from degenbot.exceptions import DegenbotError

"""
Exceptions defined here are raised by classes and functions in the `transaction` module.
"""


class TransactionError(DegenbotError):
    """
    Exception raised inside transaction simulation helpers.
    """


class DeadlineExpired(TransactionError): ...


class InsufficientOutput(TransactionError):
    def __init__(self, minimum: int, received: int) -> None:
        """
        The received amount was less than the minimum.
        """
        super().__init__(message=f"Insufficient output: {received} received, {minimum} required.")


class InsufficientInput(TransactionError):
    def __init__(self, minimum: int, deposited: int) -> None:
        """
        The deposited amount was less than the minimum.
        """
        super().__init__(message=f"Insufficient input: {deposited} deposited, {minimum} required.")


class LeftoverRouterBalance(TransactionError):
    def __init__(
        self,
        balances: dict[
            ChecksumAddress,  # token address
            int,  # balance
        ],
    ) -> None:
        self.balances = balances
        super().__init__(message="Leftover balance at router after transaction")


class PreviousBlockMismatch(TransactionError): ...


class UnknownRouterAddress(TransactionError): ...



from collections.abc import Sequence
from fractions import Fraction
from typing import Literal

import eth_abi.abi
import eth_abi.packed
from eth_typing import ChecksumAddress
from eth_utils.crypto import keccak
from hexbytes import HexBytes

from degenbot.exceptions.evm import EVMRevertError
from degenbot.functions import eip_1167_clone_address, raise_if_invalid_uint256
from degenbot.solidly.solidly_functions import (
    general_calc_d,
    general_calc_exact_in_stable,
    general_calc_k,
)


def _f_aerodrome(
    x0: int,
    y: int,
) -> int:
    _a = (x0 * y) // 10**18
    _b = (x0 * x0) // 10**18 + (y * y) // 10**18
    return (_a * _b) // 10**18


def calc_exact_in_stable(
    amount_in: int,
    token_in: Literal[0, 1],
    reserves0: int,
    reserves1: int,
    decimals0: int,
    decimals1: int,
    fee: Fraction,
) -> int:
    return general_calc_exact_in_stable(
        amount_in=amount_in,
        token_in=token_in,
        reserves0=reserves0,
        reserves1=reserves1,
        decimals0=decimals0,
        decimals1=decimals1,
        fee=fee,
        k_func=general_calc_k,
        get_y_func=_get_y_aerodrome,
    )


def _get_y_aerodrome(
    x0: int,
    xy: int,
    y: int,
    decimals0: int,
    decimals1: int,
) -> int:  # pragma: no cover
    """
    Calculate the minimum reserves for the withdrawn token that satisfy the pool invariant.

    Reference: https://github.com/aerodrome-finance/contracts/blob/main/contracts/Pool.sol
    """

    for _ in range(255):
        k = _f_aerodrome(x0, y)
        if k < xy:
            dy = ((xy - k) * 10**18) // general_calc_d(x0, y)
            if dy == 0:
                if k == xy:
                    return y
                if (
                    _k_aerodrome(
                        balance_0=x0, balance_1=y + 1, decimals_0=decimals0, decimals_1=decimals1
                    )
                    > xy
                ):
                    return y + 1
                dy = 1
            y = y + dy
        else:
            dy = ((k - xy) * 10**18) // general_calc_d(x0, y)
            if dy == 0:
                if k == xy or _f_aerodrome(x0, y - 1) < xy:
                    return y
                dy = 1
            y = y - dy
    raise EVMRevertError(error="Failed to converge on a value for y")


def _k_aerodrome(
    balance_0: int,
    balance_1: int,
    decimals_0: int,
    decimals_1: int,
) -> int:
    _x = balance_0 * 10**18 // decimals_0
    _y = balance_1 * 10**18 // decimals_1
    _a = (_x * _y) // 10**18
    _b = (_x * _x) // 10**18 + (_y * _y) // 10**18
    raise_if_invalid_uint256(_a * _b)
    return _a * _b // 10**18  # x^3*y + y^3*x >= k


def generate_aerodrome_v2_pool_address(
    *,
    deployer_address: str | bytes,
    token_addresses: Sequence[str | bytes],
    implementation_address: str | bytes,
    stable: bool,
) -> ChecksumAddress:
    """
    Get the deterministic V2 pool address generated by CREATE2. Uses the token address to generate
    the salt. The token addresses can be passed in any order.

    Adapted from https://github.com/aerodrome-finance/contracts/blob/main/contracts/factories/PoolFactory.sol
    and https://github.com/OpenZeppelin/openzeppelin-contracts/blob/master/contracts/proxy/Clones.sol
    """

    sorted_token_addresses = sorted([HexBytes(address) for address in token_addresses])

    salt = keccak(
        eth_abi.packed.encode_packed(
            ("address", "address", "bool"),
            [*sorted_token_addresses, stable],
        )
    )

    return eip_1167_clone_address(
        deployer=deployer_address,
        implementation_contract=implementation_address,
        salt=salt,
    )


def generate_aerodrome_v3_pool_address(
    deployer_address: str | bytes,
    token_addresses: Sequence[str | bytes],
    implementation_address: str | bytes,
    tick_spacing: int,
) -> ChecksumAddress:
    """
    Get the deterministic V3 pool address generated by CREATE2. Uses the token address to generate
    the salt. The token addresses can be passed in any order.

    Adapted from https://github.com/aerodrome-finance/slipstream/blob/main/contracts/core/CLFactory.sol
    and https://github.com/OpenZeppelin/openzeppelin-contracts/blob/master/contracts/proxy/Clones.sol
    """

    sorted_token_addresses = sorted([HexBytes(address) for address in token_addresses])

    salt = keccak(
        eth_abi.abi.encode(
            ("address", "address", "int24"),
            [*sorted_token_addresses, tick_spacing],
        )
    )

    return eip_1167_clone_address(
        deployer=deployer_address,
        implementation_contract=implementation_address,
        salt=salt,
    )



import dataclasses

from eth_typing import HexAddress

from degenbot.types.abstract import AbstractPoolState
from degenbot.types.concrete import PoolStateMessage


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class CurveStableswapPoolState(AbstractPoolState):
    balances: tuple[int, ...]
    base: "CurveStableswapPoolState | None" = None


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableswapPoolSimulationResult:
    amount0_delta: int
    amount1_delta: int
    current_state: CurveStableswapPoolState
    future_state: CurveStableswapPoolState


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableSwapPoolAttributes:
    address: HexAddress
    lp_token_address: HexAddress
    coin_addresses: list[HexAddress]
    coin_index_type: str
    is_metapool: bool
    underlying_coin_addresses: list[HexAddress] | None = dataclasses.field(default=None)
    base_pool_address: HexAddress | None = dataclasses.field(default=None)


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableSwapPoolStateUpdated(PoolStateMessage):
    state: CurveStableswapPoolState



import dataclasses
from bisect import bisect_left
from fractions import Fraction
from threading import Lock
from typing import TYPE_CHECKING, Any, cast
from weakref import WeakSet

import eth_abi.abi
from eth_typing import ChecksumAddress
from web3 import Web3
from web3.types import BlockIdentifier

from degenbot import connection_manager, get_checksum_address
from degenbot.aerodrome.functions import (
    calc_exact_in_stable,
    generate_aerodrome_v2_pool_address,
    generate_aerodrome_v3_pool_address,
)
from degenbot.aerodrome.types import (
    AerodromeV2PoolExternalUpdate,
    AerodromeV2PoolState,
    AerodromeV2PoolStateUpdated,
    AerodromeV3PoolState,
)
from degenbot.erc20 import Erc20Token, Erc20TokenManager
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.liquidity_pool import (
    AddressMismatch,
    ExternalUpdateError,
    InvalidSwapInputAmount,
    LateUpdateError,
    LiquidityPoolError,
    NoPoolStateAvailable,
)
from degenbot.functions import encode_function_calldata, get_number_for_block_identifier, raw_call
from degenbot.logging import logger
from degenbot.registry import pool_registry
from degenbot.solidly.solidly_functions import general_calc_exact_in_volatile
from degenbot.types.abstract import AbstractLiquidityPool
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import (
    AbstractPublisherMessage,
    BoundedCache,
    Publisher,
    PublisherMixin,
    Subscriber,
)
from degenbot.uniswap.types import UniswapPoolSwapVector
from degenbot.uniswap.v2_functions import constant_product_calc_exact_out
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool

if TYPE_CHECKING:
    from hexbytes import HexBytes


class AerodromeV2Pool(PublisherMixin, AbstractLiquidityPool):
    type PoolState = AerodromeV2PoolState
    _state_cache: BoundedCache[BlockNumber, PoolState]
    _state: PoolState

    FEE_DENOMINATOR = 10_000

    def __init__(
        self,
        address: ChecksumAddress | str,
        *,
        chain_id: ChainId | None = None,
        deployer_address: str | None = None,
        state_block: BlockNumber | None = None,
        verify_address: bool = True,
        silent: bool = False,
        state_cache_depth: int = 8,
    ) -> None:
        self.address = get_checksum_address(address)

        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id
        w3 = connection_manager.get_web3(self.chain_id)
        state_block = state_block if state_block is not None else w3.eth.block_number

        self.factory, (token0, token1), self.stable, fee, (reserves0, reserves1) = (
            self.get_factory_tokens_stable_reserves_batched(w3=w3, state_block=state_block)
        )
        self.deployer_address = (
            get_checksum_address(deployer_address) if deployer_address is not None else self.factory
        )

        self._state_lock = Lock()
        self._state = self.PoolState.__value__(
            address=self.address,
            reserves_token0=reserves0,
            reserves_token1=reserves1,
            block=state_block,
        )

        self.fee = self.fee_token0 = self.fee_token1 = Fraction(fee, type(self).FEE_DENOMINATOR)

        token_manager = Erc20TokenManager(chain_id=self.chain_id)
        self.token0, self.token1 = (
            token_manager.get_erc20token(
                address=token0,
                silent=silent,
            ),
            token_manager.get_erc20token(
                address=token1,
                silent=silent,
            ),
        )

        if verify_address and self.address != self._verified_address():  # pragma: no cover
            raise AddressMismatch

        self.name = f"{self.token0}-{self.token1} ({self.__class__.__name__}, {100 * self.fee.numerator / self.fee.denominator:.2f}%)"  # noqa:E501

        self._state_cache = BoundedCache(max_items=state_cache_depth)
        self._state_cache[self.update_block] = self.state

        pool_registry.add(pool_address=self.address, chain_id=self.chain_id, pool=self)

        self._subscribers: WeakSet[Subscriber] = WeakSet()

        if not silent:  # pragma: no cover
            logger.info(self.name)
            logger.info(f"• Token 0: {self.token0} - Reserves: {self.reserves_token0}")
            logger.info(f"• Token 1: {self.token1} - Reserves: {self.reserves_token1}")

    def __getstate__(self) -> dict[str, Any]:
        # Remove objects that either cannot be pickled or are unnecessary to perform the calculation
        copied_attributes = ()
        dropped_attributes = (
            "_state_lock",
            "_state_cache",
            "_subscribers",
        )

        with self._state_lock:
            return {
                k: (v.copy() if k in copied_attributes else v)
                for k, v in self.__dict__.items()
                if k not in dropped_attributes
            }

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(address={self.address}, token0={self.token0}, token1={self.token1}, stable={self.stable})"  # noqa:E501

    def _notify_subscribers(self: Publisher, message: AbstractPublisherMessage) -> None:
        for subscriber in self._subscribers:
            subscriber.notify(publisher=self, message=message)

    def _verified_address(self) -> ChecksumAddress:
        # The implementation address is hard-coded into the contract
        implementation_address = get_checksum_address(
            connection_manager.get_web3(self.chain_id).eth.get_code(self.address)[10:30]
        )

        return generate_aerodrome_v2_pool_address(
            deployer_address=self.deployer_address,
            token_addresses=(self.token0.address, self.token1.address),
            implementation_address=get_checksum_address(implementation_address),
            stable=self.stable,
        )

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def reserves_token0(self) -> int:
        return self.state.reserves_token0

    @property
    def reserves_token1(self) -> int:
        return self.state.reserves_token1

    @property
    def state(self) -> PoolState:
        return self._state

    @property
    def tokens(self) -> tuple[Erc20Token, Erc20Token]:
        return self.token0, self.token1

    @property
    def update_block(self) -> BlockNumber:
        if TYPE_CHECKING:
            assert self.state.block is not None
        return self.state.block

    @property
    def w3(self) -> Web3:
        return connection_manager.get_web3(self.chain_id)

    def swap_is_viable(
        self,
        state: PoolState,
        vector: UniswapPoolSwapVector,
    ) -> bool:
        if state.reserves_token0 == 0 or state.reserves_token1 == 0:
            return False
        return state.reserves_token1 > 1 if vector.zero_for_one else state.reserves_token0 > 1

    def auto_update(
        self,
        *,
        block_number: BlockNumber | None = None,
        silent: bool = True,
    ) -> None:
        """
        Retrieves and records the current state from the pool at the provided block number, or the
        latest block if not provided.

        @dev this method uses a lock to guard state-modifying methods that might cause race
        conditions when used with threads.
        """
        with self._state_lock:
            if block_number is not None and block_number < self.update_block:
                raise LateUpdateError

            state_updated = False
            w3 = self.w3
            block_number = block_number if block_number is not None else w3.eth.get_block_number()

            reserves0, reserves1 = self.get_reserves(w3=w3, block_identifier=block_number)

            if (self.reserves_token0, self.reserves_token1) != (reserves0, reserves1):
                state_updated = True
                self._state = dataclasses.replace(
                    self.state,
                    reserves_token0=reserves0,
                    reserves_token1=reserves1,
                    block=block_number,
                )

            if state_updated:
                self._state_cache[block_number] = self.state
                self._notify_subscribers(
                    message=AerodromeV2PoolStateUpdated(self.state),
                )

                if not silent:  # pragma: no cover
                    logger.info(f"[{self.name}]")
                    logger.info(f"{self.token0}: {self.reserves_token0}")
                    logger.info(f"{self.token1}: {self.reserves_token1}")

    def calculate_tokens_in_from_tokens_out(
        self,
        token_out_quantity: int,
        token_out: Erc20Token,
        override_state: PoolState | None = None,
    ) -> int:
        """
        Calculates the required token INPUT of token_in for a target OUTPUT at current pool
        reserves.

        Accepts a `PoolState` state override for calculation against an arbitrary state
        in lieu of the recorded state.
        """

        if token_out_quantity <= 0:  # pragma: no cover
            raise InvalidSwapInputAmount

        if override_state:  # pragma: no cover
            logger.debug(f"State overrides applied: {override_state}")

        if token_out == self.token1:
            reserves_in = (
                override_state.reserves_token0
                if override_state is not None
                else self.reserves_token0
            )
            reserves_out = (
                override_state.reserves_token1
                if override_state is not None
                else self.reserves_token1
            )

        elif token_out == self.token0:
            reserves_in = (
                override_state.reserves_token1
                if override_state is not None
                else self.reserves_token1
            )
            reserves_out = (
                override_state.reserves_token0
                if override_state is not None
                else self.reserves_token0
            )

        else:  # pragma: no cover
            raise DegenbotValueError(
                message=f"Could not identify token_out: {token_out}! This pool holds: {self.token0} {self.token1}"  # noqa:E501
            )

        # last token becomes infinitely expensive, so largest possible swap out is reserves - 1
        if token_out_quantity > reserves_out - 1:
            raise LiquidityPoolError(
                message=f"Requested amount out ({token_out_quantity}) >= pool reserves ({reserves_out})"  # noqa:E501
            )

        if self.stable:
            raise NotImplementedError

        return constant_product_calc_exact_out(
            amount_out=token_out_quantity,
            reserves_in=reserves_in,
            reserves_out=reserves_out,
            fee=self.fee,
        )

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        override_state: PoolState | None = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT for a target INPUT at current pool reserves.
        """

        if token_in not in self.tokens:  # pragma: no cover
            raise DegenbotValueError(message="token_in not recognized.")

        if token_in_quantity <= 0:  # pragma: no cover
            raise InvalidSwapInputAmount

        if override_state:  # pragma: no cover
            logger.debug(f"State overrides applied: {override_state}")

        reserves_0 = (
            override_state.reserves_token0 if override_state is not None else self.reserves_token0
        )
        reserves_1 = (
            override_state.reserves_token1 if override_state is not None else self.reserves_token1
        )

        if self.stable:
            return calc_exact_in_stable(
                amount_in=token_in_quantity,
                token_in=0 if token_in == self.token0 else 1,
                reserves0=reserves_0,
                reserves1=reserves_1,
                decimals0=10**self.token0.decimals,
                decimals1=10**self.token1.decimals,
                fee=self.fee,
            )
        return general_calc_exact_in_volatile(
            amount_in=token_in_quantity,
            token_in=0 if token_in == self.token0 else 1,
            reserves0=reserves_0,
            reserves1=reserves_1,
            fee=self.fee,
        )

    def external_update(
        self,
        update: AerodromeV2PoolExternalUpdate,
    ) -> None:
        if update.block_number < self.update_block:
            raise ExternalUpdateError(
                message=f"Rejected update for block {update.block_number} in the past, current update block is {self.update_block}"  # noqa:E501
            )

        with self._state_lock:
            self._state = dataclasses.replace(
                self.state,
                reserves_token0=update.reserves_token0,
                reserves_token1=update.reserves_token1,
                block=update.block_number,
            )
            self._state_cache[update.block_number] = self.state
            self._notify_subscribers(
                message=AerodromeV2PoolStateUpdated(self.state),
            )

    def get_absolute_price(
        self, token: Erc20Token, override_state: PoolState | None = None
    ) -> Fraction:
        """
        Get the absolute price for the given token, expressed in units of the other.
        """

        return 1 / self.get_absolute_exchange_rate(token, override_state=override_state)

    def get_absolute_exchange_rate(
        self,
        token: Erc20Token,
        override_state: PoolState | None = None,
    ) -> Fraction:
        """
        Get the absolute exchange rate for the given token, expressed in terms of a unit amount of
        its paired token.

        e.g. taking the USDC-WETH pool in https://blog.uniswap.org/uniswap-v3-math-primer — the
        WETH/USDC exchange rate is 649004842.70137. Rounding down, this signifies that the smallest
        swap (1 USDC) results in a 649004842 WETH output.

        The exchange rate for a V2 pool is a simple ratio of the output token reserves to the input
        token reserves.
        """

        if token not in self.tokens:
            raise DegenbotValueError(message=f"Unknown token {token}")

        state = self.state if override_state is None else override_state

        return (
            Fraction(state.reserves_token1, state.reserves_token0)
            if token == self.token1
            else Fraction(state.reserves_token0, state.reserves_token1)
        )

    def get_nominal_price(
        self,
        token: Erc20Token,
        override_state: PoolState | None = None,
    ) -> Fraction:
        """
        Get the nominal price for the given token, expressed per nominal unit of its paired token.
        The price is corrected for the decimal place values of both tokens.
        """

        return 1 / self.get_nominal_exchange_rate(token=token, override_state=override_state)

    def get_nominal_exchange_rate(
        self,
        token: Erc20Token,
        override_state: PoolState | None = None,
    ) -> Fraction:
        """
        Get the nominal rate for the given token, expressed in units of the other, corrected for
        decimal place values.
        """

        return self.get_absolute_exchange_rate(token=token, override_state=override_state) * (
            Fraction(10**self.token1.decimals, 10**self.token0.decimals)
            if token == self.token0
            else Fraction(10**self.token0.decimals, 10**self.token1.decimals)
        )

    def get_factory_tokens_stable_reserves_batched(
        self,
        w3: Web3,
        state_block: BlockNumber,
    ) -> tuple[
        ChecksumAddress,  # factory
        tuple[ChecksumAddress, ChecksumAddress],  # tokens
        bool,  # stable
        int,  # fee
        tuple[int, int],  # reserves
    ]:
        with w3.batch_requests() as batch:
            batch.add_mapping(
                {
                    # These calls default to use 'latest' for block number, which is OK since the
                    # values are immutable
                    w3.eth.call: [
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="factory()",
                                function_arguments=None,
                            ),
                        },
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="token0()",
                                function_arguments=None,
                            ),
                        },
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="token1()",
                                function_arguments=None,
                            ),
                        },
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="stable()",
                                function_arguments=None,
                            ),
                        },
                    ],
                }
            )
            batch.add(
                # This call uses a specific block so the reserve values are consistent
                w3.eth.call(
                    transaction={
                        "to": self.address,
                        "data": encode_function_calldata(
                            function_prototype="getReserves()",
                            function_arguments=None,
                        ),
                    },
                    block_identifier=state_block,
                )
            )

            factory, token0, token1, stable, reserves = batch.execute()

        (factory,) = eth_abi.abi.decode(types=["address"], data=cast("HexBytes", factory))
        (token0,) = eth_abi.abi.decode(types=["address"], data=cast("HexBytes", token0))
        (token1,) = eth_abi.abi.decode(types=["address"], data=cast("HexBytes", token1))
        (stable,) = eth_abi.abi.decode(types=["bool"], data=cast("HexBytes", stable))
        reserves0, reserves1, _ = eth_abi.abi.decode(
            types=["uint256", "uint256", "uint256"], data=cast("HexBytes", reserves)
        )

        (fee,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": get_checksum_address(cast("str", factory)),
                    "data": encode_function_calldata(
                        function_prototype="getFee(address,bool)",
                        function_arguments=[self.address, stable],
                    ),
                }
            ),
        )

        return (
            get_checksum_address(cast("str", factory)),
            (get_checksum_address(cast("str", token0)), get_checksum_address(cast("str", token1))),
            cast("bool", stable),
            cast("int", fee),
            (cast("int", reserves0), cast("int", reserves1)),
        )

    def get_reserves(
        self, w3: Web3, block_identifier: BlockIdentifier | None = None
    ) -> tuple[int, int]:
        reserves_token0, reserves_token1 = raw_call(
            w3=w3,
            address=self.address,
            block_identifier=get_number_for_block_identifier(block_identifier, w3),
            calldata=encode_function_calldata(
                function_prototype="getReserves()",
                function_arguments=None,
            ),
            return_types=["uint256", "uint256"],
        )

        return cast("int", reserves_token0), cast("int", reserves_token1)

    def discard_states_before_block(self, block: BlockNumber) -> None:
        """
        Discard states recorded prior to a target block.
        """

        with self._state_lock:
            known_blocks = sorted(self._state_cache.keys())

            # Finds the index prior to the requested block number
            block_index = bisect_left(known_blocks, block)

            # The earliest known state already meets the criterion, so return early
            if block_index == 0:
                return

            if block_index == len(known_blocks):
                raise NoPoolStateAvailable(block=block)

            for known_block in known_blocks[:block_index]:
                del self._state_cache[known_block]

    def restore_state_before_block(
        self,
        block: BlockNumber,
    ) -> None:
        """
        Restore the last pool state recorded prior to a target block.

        Use this method to maintain consistent state data following a chain re-organization.
        """

        with self._state_lock:
            known_blocks = sorted(self._state_cache.keys())

            # Finds the index prior to the requested block number
            block_index = bisect_left(known_blocks, block)

            if block_index == 0:
                raise NoPoolStateAvailable(block=block)

            # The last known state already meets the criterion, so return early
            if block_index == len(known_blocks):
                return

            # Remove states at and after the specified block
            for known_block in known_blocks[block_index:]:
                del self._state_cache[known_block]

            # Restore previous state and block
            self._state = list(self._state_cache.values())[-1]
            self._notify_subscribers(message=AerodromeV2PoolStateUpdated(self.state))


class AerodromeV3Pool(UniswapV3Pool):
    type PoolState = AerodromeV3PoolState

    TICK_STRUCT_TYPES = (
        "uint128",
        "int128",
        "int128",
        "uint256",
        "uint256",
        "uint256",
        "int56",
        "uint160",
        "uint32",
        "bool",
    )  # type:ignore[assignment]

    SLOT0_STRUCT_TYPES = (
        "uint160",
        "int24",
        "uint16",
        "uint16",
        "uint16",
        "bool",
    )  # type:ignore[assignment]

    def _verified_address(self) -> ChecksumAddress:
        # The implementation address is hard-coded into the contract
        implementation_address = get_checksum_address(
            connection_manager.get_web3(self.chain_id).eth.get_code(self.address)[10:30]
        )

        return generate_aerodrome_v3_pool_address(
            deployer_address=self.deployer_address,
            token_addresses=(self.token0.address, self.token1.address),
            implementation_address=get_checksum_address(implementation_address),
            tick_spacing=self.tick_spacing,
        )



from degenbot import AnvilFork, ChainlinkPriceContract, get_checksum_address, set_web3


def test_chainlink_feed(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)

    # Load WETH price feed
    # ref: https://data.chain.link/ethereum/mainnet/crypto-usd/eth-usd
    weth_price_feed = ChainlinkPriceContract(
        get_checksum_address("0x5f4ec3df9cbd43714fe2740f5e3616155c5b8419")
    )
    assert isinstance(weth_price_feed.price, float)



import pickle
from typing import Any

from degenbot.types.concrete import BoundedCache


def test_pickle_bounded_cache():
    bounded_cache: BoundedCache[Any, Any] = BoundedCache(max_items=5)
    pickled_cache = pickle.dumps(bounded_cache)
    pickle.loads(pickled_cache)


def test_bounded_cache_max_items():
    bounded_cache: BoundedCache[Any, Any] = BoundedCache(max_items=5)
    bounded_cache[0] = "a"
    bounded_cache[1] = "b"
    bounded_cache[2] = "c"
    bounded_cache[3] = "d"
    bounded_cache[4] = "e"
    assert len(bounded_cache) == 5
    assert bounded_cache.keys() == {0, 1, 2, 3, 4}

    # Adding this element should cause the first to be removed
    bounded_cache[5] = "f"
    assert len(bounded_cache) == 5
    assert bounded_cache.keys() == {1, 2, 3, 4, 5}

    # This element already exists, so it should update the value
    bounded_cache[2] = "c"
    assert len(bounded_cache) == 5
    assert bounded_cache.keys() == {1, 2, 3, 4, 5}



import dataclasses

from degenbot.types.abstract import AbstractPoolState, AbstractSimulationResult
from degenbot.types.aliases import BlockNumber
from degenbot.types.concrete import PoolStateMessage


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapSimulationResult(AbstractSimulationResult):
    """
    Common attributes for Uniswap V2 & V3 simulations
    """

    amount0_delta: int
    amount1_delta: int
    initial_state: AbstractPoolState
    final_state: AbstractPoolState


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class UniswapV2PoolState(AbstractPoolState):
    reserves_token0: int
    reserves_token1: int


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV2PoolSimulationResult(UniswapSimulationResult):
    initial_state: UniswapV2PoolState
    final_state: UniswapV2PoolState


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV2PoolExternalUpdate:
    block_number: BlockNumber
    reserves_token0: int
    reserves_token1: int


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV2PoolStateUpdated(PoolStateMessage):
    state: UniswapV2PoolState



from . import abi
from .managers import AerodromeV2PoolManager, AerodromeV3PoolManager
from .pools import AerodromeV2Pool, AerodromeV3Pool
from .types import AerodromeV2PoolState, AerodromeV3PoolState

__all__ = (
    "AerodromeV2Pool",
    "AerodromeV2PoolManager",
    "AerodromeV2PoolState",
    "AerodromeV3Pool",
    "AerodromeV3PoolManager",
    "AerodromeV3PoolState",
    "abi",
)



from eth_typing import HexStr
from hexbytes import HexBytes

MAX_LITERALS = 32
MAX_MATCH_LENGTH = 262
MAX_MATCH_OFFSET = 8191

MAX_1BYTE_INT = 0xFF
MAX_3BYTE_INT = 0xFFFFFF


def flz_compress(uncompressed_data: str | bytes) -> HexBytes:
    """
    Compress data using Solady's FastLZ algorithm.

    ref: https://github.com/Vectorized/solady/blob/main/js/solady.js
    """

    def get_3_byte_int(pos: int) -> int:
        """
        Get a 24-bit (3 byte) unsigned integer from the input buffer at position `pos`.

        The number is constructed from a little-endian memory arrangement (the least significant
        digit appears first).
        """

        return input_buffer[pos] + (input_buffer[pos + 1] << 8) + (input_buffer[pos + 2] << 16)

    def hash_(x: int) -> int:
        """
        Hash the 1-byte value
        """

        return ((2654435769 * x) >> 19) & (MAX_MATCH_OFFSET)

    def add_literals(run_length: int, input_buffer_start_offset: int) -> None:
        """
        Extend the output buffer with a continuous run of 1-byte literals copied from the input
        buffer.
        """

        while run_length > 0:
            number_of_literals = min(run_length, MAX_LITERALS)

            # Encode the run length
            output_buffer.append(number_of_literals - 1)

            # Encode the literals
            output_buffer.extend(
                input_buffer[
                    input_buffer_start_offset : input_buffer_start_offset + number_of_literals
                ]
            )

            run_length -= number_of_literals
            input_buffer_start_offset += number_of_literals

    input_buffer = bytearray(HexBytes(uncompressed_data))
    compressible_bytes = len(input_buffer) - 4
    hash_table: dict[int, int] = {}
    output_buffer = bytearray()

    last_literal_offset = 0
    input_buffer_offset = 2

    while input_buffer_offset < compressible_bytes - 9:
        while True:
            input_chunk = get_3_byte_int(input_buffer_offset)
            input_chunk_hash = hash_(input_chunk)

            last_seen_offset = hash_table.get(input_chunk_hash, 0)

            # Always update the last-seen offset for the chunk
            hash_table[input_chunk_hash] = input_buffer_offset

            # Measure the distance from the last-seen chunk, or the start of the buffer
            # (if this is the first occurence)
            distance_to_last_seen_chunk = input_buffer_offset - last_seen_offset
            reference_chunk = (
                get_3_byte_int(last_seen_offset)
                if distance_to_last_seen_chunk <= MAX_MATCH_OFFSET
                else MAX_3BYTE_INT + 1
            )

            if input_buffer_offset < compressible_bytes - 9 and input_chunk != reference_chunk:
                input_buffer_offset += 1
                continue

            input_buffer_offset += 1
            break

        if input_buffer_offset >= compressible_bytes - 9:
            break

        input_buffer_offset -= 1
        if input_buffer_offset > last_literal_offset:
            add_literals(
                run_length=input_buffer_offset - last_literal_offset,
                input_buffer_start_offset=last_literal_offset,
            )

        p = last_seen_offset + 3  # Start of previous occurrence
        q = input_buffer_offset + 3  # Start of current occurrence
        max_possible_match_length = compressible_bytes - q
        match_length = max_possible_match_length

        # Find maximum match length
        for offset in range(max_possible_match_length):
            if input_buffer[p + offset] != input_buffer[q + offset]:
                match_length = offset + 1
                break

        input_buffer_offset += match_length
        distance_to_last_seen_chunk -= 1  # Adjust offset for encoding

        while match_length >= MAX_MATCH_LENGTH:
            # LONG MATCH instruction - 3 byte opcode
            opcode_0 = 0b11100000 + (distance_to_last_seen_chunk >> 8)
            opcode_1 = MAX_MATCH_LENGTH - 9
            opcode_2 = distance_to_last_seen_chunk & 0b0000011111111

            output_buffer.append(opcode_0)
            output_buffer.append(opcode_1)
            output_buffer.append(opcode_2)

            match_length -= MAX_MATCH_LENGTH

        # Encode the remaining chunk
        if match_length < 7:  # noqa: PLR2004
            # SHORT MATCH instruction - 2 byte opcode
            opcode_0 = (match_length << 5) + (distance_to_last_seen_chunk >> 8)
            opcode_1 = distance_to_last_seen_chunk & 0b11111111

            output_buffer.append(opcode_0)
            output_buffer.append(opcode_1)
        else:
            # LONG MATCH instruction - 3 byte opcode
            opcode_0 = 0b11100000 + (distance_to_last_seen_chunk >> 8)
            opcode_1 = match_length - 7
            opcode_2 = distance_to_last_seen_chunk & 0b0000011111111

            output_buffer.append(opcode_0)
            output_buffer.append(opcode_1)
            output_buffer.append(opcode_2)

        # Update hash table with next positions
        hash_table[hash_(get_3_byte_int(input_buffer_offset))] = input_buffer_offset
        input_buffer_offset += 1
        hash_table[hash_(get_3_byte_int(input_buffer_offset))] = input_buffer_offset
        input_buffer_offset += 1

        # Update last literals position
        last_literal_offset = input_buffer_offset

    # Add remaining literals
    add_literals(
        run_length=compressible_bytes + 4 - last_literal_offset,
        input_buffer_start_offset=last_literal_offset,
    )

    return HexBytes(output_buffer)


def flz_decompress(compressed_data: bytes | bytearray | HexStr) -> HexBytes:
    """
    Decompress data using Solady's FastLZ algorithm.

    ref: https://github.com/Vectorized/solady/blob/main/js/solady.js
    """

    input_buffer = bytearray(HexBytes(compressed_data))
    output_buffer = bytearray()

    while input_buffer:
        opcode_0 = input_buffer.pop(0)
        instruction_type = opcode_0 >> 5

        match instruction_type:
            case 0:  # Literal run
                number_of_literals = 1 + (opcode_0 & 0b00011111)
                for _ in range(number_of_literals):
                    output_buffer.append(input_buffer.pop(0))

            case 1 | 2 | 3 | 4 | 5 | 6:  # Short match
                match_length = 2 + (opcode_0 >> 5)

                opcode_1 = input_buffer.pop(0)
                reference_offset = 256 * (opcode_0 & 0b00011111) + opcode_1

                for _ in range(match_length):
                    output_buffer.append(output_buffer[-1 - reference_offset])

            case 7:  # Long match
                opcode_1 = input_buffer.pop(0)
                match_length = 9 + opcode_1

                opcode_2 = input_buffer.pop(0)
                reference_offset = 256 * (opcode_0 & 0b00011111) + opcode_2

                for _ in range(match_length):
                    output_buffer.append(output_buffer[-1 - reference_offset])

            case _:
                error_message = "Invalid instruction!"
                raise ValueError(error_message)

    return HexBytes(output_buffer)



import pathlib
import tomllib
from typing import Annotated

import tomlkit
from pydantic import BaseModel, FilePath, PlainSerializer
from pydantic_settings import BaseSettings, SettingsConfigDict

CONFIG_DIR = pathlib.Path.home() / ".config" / "degenbot"
CONFIG_FILE = CONFIG_DIR / "config.toml"
DEFAULT_DB_PATH = CONFIG_DIR / "degenbot.db"


class DatabaseSettings(BaseModel):
    # Serialize the path as a string representation of the absolute path
    path: Annotated[
        FilePath,
        PlainSerializer(lambda path: str(path.absolute()), return_type=str),
    ]


class Settings(BaseSettings):
    model_config = SettingsConfigDict()

    database: DatabaseSettings


def load_config_from_file(config_path: pathlib.Path) -> Settings:
    return Settings.model_validate(
        tomllib.loads(
            config_path.read_text(),
        ),
    )


def save_config_to_file(config: Settings) -> None:
    if not CONFIG_DIR.exists():
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    CONFIG_FILE.write_text(
        tomlkit.dumps(
            config.model_dump(),
        ),
    )


if not CONFIG_DIR.exists():
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)

if not CONFIG_FILE.exists():
    from degenbot.database import create_new_sqlite_database

    # Pydantic will validate that the DB path exists, so initialize it first
    create_new_sqlite_database(DEFAULT_DB_PATH)

    _default_settings = Settings(
        database=DatabaseSettings(
            path=DEFAULT_DB_PATH,
        ),
    )

    save_config_to_file(_default_settings)

settings: Settings = load_config_from_file(CONFIG_FILE)



from fractions import Fraction
from threading import Lock

import eth_abi.abi
from eth_typing import ChecksumAddress

from degenbot import connection_manager, get_checksum_address
from degenbot.balancer.libraries.fixed_point import mul_up
from degenbot.balancer.libraries.scaling_helpers import (
    _compute_scaling_factor,
    _downscale_down,
    _upscale,
    _upscale_array,
)
from degenbot.balancer.libraries.weighted_math import _calc_out_given_in, _subtract_swap_fee_amount
from degenbot.balancer.types import BalancerV2PoolState
from degenbot.erc20 import Erc20Token, Erc20TokenManager
from degenbot.functions import encode_function_calldata
from degenbot.types.abstract import AbstractLiquidityPool
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import PublisherMixin


class BalancerV2Pool(PublisherMixin, AbstractLiquidityPool):
    type PoolState = BalancerV2PoolState
    FEE_DENOMINATOR = 1 * 10**18

    def __init__(
        self,
        address: ChecksumAddress | str,
        *,
        chain_id: ChainId | None = None,
        state_block: BlockNumber | None = None,
        verify_address: bool = False,
        silent: bool = False,
    ) -> None:
        self.address = get_checksum_address(address)

        if verify_address:
            # TODO: add functionality
            raise NotImplementedError

        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id
        w3 = connection_manager.get_web3(self.chain_id)
        state_block = state_block if state_block is not None else w3.eth.block_number

        pool_id: bytes
        (pool_id,) = eth_abi.abi.decode(
            types=["bytes32"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": encode_function_calldata(
                        function_prototype="getPoolId()",
                        function_arguments=None,
                    ),
                },
                block_identifier=state_block,
            ),
        )
        self.pool_id = pool_id
        self.pool_specialization = int.from_bytes(self.pool_id[20:22], byteorder="big")

        vault_address: str
        (vault_address,) = eth_abi.abi.decode(
            types=["address"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": encode_function_calldata(
                        function_prototype="getVault()",
                        function_arguments=None,
                    ),
                },
                block_identifier=state_block,
            ),
        )
        self.vault = get_checksum_address(vault_address)

        tokens, balances, _ = eth_abi.abi.decode(
            types=["address[]", "uint256[]", "uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.vault,
                    "data": encode_function_calldata(
                        function_prototype="getPoolTokens(bytes32)",
                        function_arguments=[self.pool_id],
                    ),
                },
                block_identifier=state_block,
            ),
        )

        token_manager = Erc20TokenManager(chain_id=self.chain_id)
        self.tokens = tuple(
            token_manager.get_erc20token(
                address=get_checksum_address(token),
                silent=silent,
            )
            for token in tokens
        )
        self.scaling_factors = tuple([_compute_scaling_factor(token) for token in self.tokens])

        self._state_lock = Lock()
        self._state = BalancerV2PoolState(
            address=self.address,
            block=state_block,
            balances=tuple(balances),
        )

        (fee,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": encode_function_calldata(
                        function_prototype="getSwapFeePercentage()",
                        function_arguments=None,
                    ),
                },
                block_identifier=state_block,
            ),
        )
        self.fee = Fraction(fee, self.FEE_DENOMINATOR)

        (weights,) = eth_abi.abi.decode(
            types=["uint256[]"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": encode_function_calldata(
                        function_prototype="getNormalizedWeights()",
                        function_arguments=None,
                    ),
                },
                block_identifier=state_block,
            ),
        )
        self.weights = tuple(weights)

    @property
    def balances(self) -> tuple[int, ...]:
        return self.state.balances

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def state(self) -> PoolState:
        return self._state

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        token_out: Erc20Token,
        override_state: PoolState | None = None,
    ) -> int:
        token_in_index = self.tokens.index(token_in)
        token_out_index = self.tokens.index(token_out)

        fee_amount = mul_up(token_in_quantity, self.fee * self.FEE_DENOMINATOR)

        amount_new = _subtract_swap_fee_amount(
            amount=token_in_quantity,
            fee_percentage=self.fee * self.FEE_DENOMINATOR,
        )

        assert token_in_quantity - fee_amount == amount_new

        if override_state is not None:
            # TODO: add functionality
            raise NotImplementedError
        balances = list(self.balances)  # make a copy because _upscale_array will mutate it

        _upscale_array(amounts=balances, scaling_factors=self.scaling_factors)
        amount_new = _upscale(amount_new, scaling_factor=self.scaling_factors[token_in_index])

        amount_out = _calc_out_given_in(
            balance_in=int(balances[token_in_index]),
            weight_in=self.weights[token_in_index],
            balance_out=int(balances[token_out_index]),
            weight_out=self.weights[token_out_index],
            amount_in=int(amount_new),
        )

        return int(
            _downscale_down(amount=amount_out, scaling_factor=self.scaling_factors[token_out_index])
        )



import contextlib
import pathlib
import shutil
import socket
import subprocess
import tempfile
from collections.abc import AsyncIterator, Iterable
from typing import TYPE_CHECKING, Any, Literal, cast

import tenacity
from eth_typing import HexAddress, HexStr
from hexbytes import HexBytes
from pydantic import validate_call
from web3 import AsyncIPCProvider, AsyncWeb3, IPCProvider, Web3
from web3.middleware import Middleware
from web3.types import RPCEndpoint

from degenbot.exceptions import DegenbotError, DegenbotValueError
from degenbot.logging import logger
from degenbot.types.aliases import BlockNumber
from degenbot.validation.evm_values import ValidatedUint256


class AnvilNotFound(Exception):
    def __init__(self) -> None:  # pragma: no cover
        super().__init__("Anvil path could not be located.")


type AnvilCommandList = list[str]


class AnvilFork:
    """
    Launch an Anvil fork as a separate process and expose methods for commonly-used RPC calls.

    Provides a `Web3` connector to Anvil's IPC socket endpoint at the `.w3` attribute.
    """

    def __init__(
        self,
        *,
        fork_url: str,
        fork_block: BlockNumber | None = None,
        fork_transaction_hash: str | None = None,
        mining_mode: Literal["auto", "interval", "none"] = "auto",
        mining_interval: int | None = None,
        storage_caching: bool = True,
        base_fee: int | None = None,
        ipc_path: pathlib.Path | None = None,
        mnemonic: str = (
            # Default mnemonic used by Brownie for Ganache forks
            "patient rude simple dog close planet oval animal hunt sketch suspect slim"
        ),
        coinbase: HexAddress | None = None,
        middlewares: list[tuple[Middleware, int]] | None = None,
        balance_overrides: Iterable[tuple[HexAddress, int]] | None = None,
        bytecode_overrides: Iterable[tuple[HexAddress, bytes]] | None = None,
        nonce_overrides: Iterable[tuple[HexAddress, int]] | None = None,
        storage_overrides: Iterable[tuple[HexAddress | bytes, int, HexStr | bytes | int]]
        | None = None,
        ipc_provider_kwargs: dict[str, Any] | None = None,
        anvil_opts: list[str] | None = None,  # Additional options passed to the Anvil command
    ) -> None:
        def _parse_base_fee_arg(command: AnvilCommandList) -> None:
            if base_fee:
                command.append(f"--base-fee={base_fee}")

        def _parse_block_number_arg(command: AnvilCommandList) -> None:
            if fork_block:
                command.append(f"--fork-block-number={fork_block}")

        def _parse_mining_mode_arg(command: AnvilCommandList) -> None:
            match mining_mode:
                case "auto":
                    return
                case "interval":
                    if mining_interval is None:
                        raise DegenbotValueError(
                            message="Interval mining mode was specified without an interval value."
                        )
                    command.append(f"--block-time={mining_interval}")
                case "none":
                    command.append("--no-mining")
                    command.append("--order=fifo")
                case _:
                    raise DegenbotValueError(message=f"Unknown mining mode '{mining_mode}'.")

        def _parse_storage_caching_arg(command: AnvilCommandList) -> None:
            if storage_caching is False:
                command.append("--no-storage-caching")

        def _parse_transaction_hash_arg(command: AnvilCommandList) -> None:
            if fork_transaction_hash:
                command.append(f"--fork-transaction-hash={fork_transaction_hash}")

        _path_to_anvil = shutil.which("anvil")
        if _path_to_anvil is None:  # pragma: no cover
            raise AnvilNotFound
        path_to_anvil = pathlib.Path(_path_to_anvil)

        self.ipc_path = (
            pathlib.Path(
                tempfile.gettempdir(),
            )
            if ipc_path is None
            else ipc_path
        )

        if ipc_provider_kwargs is not None:
            self.ipc_provider_kwargs = ipc_provider_kwargs
        else:
            self.ipc_provider_kwargs = {}

        self.port = self._get_free_port_number()

        command: AnvilCommandList = [
            str(path_to_anvil),
            "--silent",
            "--auto-impersonate",
            "--no-rate-limit",
            f"--fork-url={fork_url}",
            f"--port={self.port}",
            f"--ipc={self.ipc_filename}",
            f"--mnemonic={mnemonic}",
        ]
        _parse_base_fee_arg(command)
        _parse_block_number_arg(command)
        _parse_mining_mode_arg(command)
        _parse_storage_caching_arg(command)
        _parse_transaction_hash_arg(command)
        if anvil_opts:
            command.extend(anvil_opts)

        self._anvil_command = command
        self._setup_process(self._anvil_command)
        self._setup_w3()

        self._fork_url = fork_url

        if middlewares is not None:
            for middleware, layer in middlewares:
                self.w3.middleware_onion.inject(middleware, layer=layer)

        if balance_overrides is not None:
            for account, balance in balance_overrides:
                self.set_balance(account, balance)

        if bytecode_overrides is not None:
            for account, bytecode in bytecode_overrides:
                self.set_code(account, bytecode)

        if nonce_overrides is not None:
            for account, nonce in nonce_overrides:
                self.set_nonce(account, nonce)

        if storage_overrides is not None:
            for address, position, value in storage_overrides:
                self.set_storage(
                    address=address,
                    position=position,
                    value=value,
                )

        if coinbase is not None:
            self.set_coinbase(coinbase)

        if mining_interval:
            self.set_block_timestamp_interval(mining_interval)

    @property
    def fork_url(self) -> str:
        return self._fork_url

    @property
    def http_url(self) -> str:
        return f"http://localhost:{self.port}"

    @property
    def ipc_filename(self) -> pathlib.Path:
        return self.ipc_path / f"anvil-{self.port}.ipc"

    @property
    def ws_url(self) -> str:
        return f"ws://localhost:{self.port}"

    @staticmethod
    def _get_free_port_number() -> int:
        with socket.socket() as sock:
            sock.bind(("", 0))
            _, _port = sock.getsockname()
            return cast("int", _port)

    def _setup_w3(self) -> None:
        """
        Create a Web3 connection to the IPC socket used by Anvil, waiting for the connection to be
        established.
        """

        try:
            # network I/O is less reliable, so wait with an exponential delay and jitter
            w3 = Web3(IPCProvider(ipc_path=self.ipc_filename, **self.ipc_provider_kwargs))
            w3_connected_check_with_retry = tenacity.Retrying(
                stop=tenacity.stop_after_delay(10),
                wait=tenacity.wait_exponential_jitter(),
                retry=tenacity.retry_if_result(lambda result: result is False),
            )
            w3_connected_check_with_retry(fn=w3.is_connected)
        except tenacity.RetryError as exc:
            raise DegenbotError(message="Timed out waiting for Web3 connection.") from exc

        self.w3 = w3

    def _setup_process(self, anvil_command: AnvilCommandList) -> None:
        """
        Launch an Anvil subprocess, waiting for the IPC socket to be created.
        """

        process = subprocess.Popen(anvil_command)  # noqa: S603

        try:
            # Storage I/O should be fast, so use a low fixed wait time
            filename_check_with_retry = tenacity.Retrying(
                stop=tenacity.stop_after_delay(10),
                wait=tenacity.wait_fixed(0.01),
                retry=tenacity.retry_if_result(lambda result: result is False),
            )
            filename_check_with_retry(fn=self.ipc_filename.exists)
        except tenacity.RetryError as exc:
            raise DegenbotError(message="Timed out waiting for IPC socket to be created.") from exc

        self._process = process

    def __del__(self) -> None:
        if hasattr(self, "_process"):
            self._process.terminate()
        self.ipc_filename.unlink(missing_ok=True)

    def mine(self) -> None:
        method = "evm_mine"
        resp = self.w3.provider.make_request(
            method=RPCEndpoint(method),
            params=[],
        )
        if "error" in resp:
            raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    async def mine_async(self) -> None:
        async with self.async_w3 as async_w3:
            await async_w3.provider.make_request(
                method=RPCEndpoint("evm_mine"),
                params=[],
            )

    @property
    @contextlib.asynccontextmanager
    async def async_w3(self) -> AsyncIterator[AsyncWeb3]:
        # TODO: investigate why cache_allowed_requests causes sequential is_connected calls to fail
        async with AsyncWeb3(AsyncIPCProvider(self.ipc_filename)) as async_w3:
            if TYPE_CHECKING:
                assert isinstance(async_w3, AsyncWeb3)
            yield async_w3

    async def reset_async(
        self,
        block_number: BlockNumber,
    ) -> None:
        """
        Reset to a new block number.
        """

        method = "anvil_reset"
        async with self.async_w3 as async_w3:
            resp = await async_w3.provider.make_request(
                method=RPCEndpoint(method),
                params=[{"forking": {"blockNumber": block_number}}],
            )
            if "error" in resp:
                raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    def reset(
        self,
        fork_url: str | None = None,
        block_number: BlockNumber | None = None,
        transaction_hash: str | None = None,
    ) -> None:
        """
        Fork from a new endpoint, block number, or transaction hash.

        Resetting to a new block number only can be done in-place without relaunching the Anvil
        process or recreating the Web3 object. Resetting to a new endpoint or from a transaction
        hash will create a new Anvil process, which is slower.
        """

        if fork_url is not None or transaction_hash is not None:
            self._process.terminate()

            if block_number is not None:
                logger.warning(
                    f"Forking from transaction hash {transaction_hash}, ignoring provided block number."
                )

            # Sanitize the command by stripping options that may conflict
            self._anvil_command = [
                option
                for option in self._anvil_command.copy()
                if all(
                    (
                        "--fork-url" not in option,
                        "--fork-block-number" not in option,
                        "--fork-transaction-hash" not in option,
                    )
                )
            ]

            # Fork URL must be provided since a new process is being launched
            if fork_url is not None:
                self._fork_url = fork_url
            self._anvil_command.append(f"--fork-url={self._fork_url}")

            if block_number is not None:
                self._anvil_command.append(f"--fork-block-number={block_number}")

            if transaction_hash is not None:
                self._anvil_command.append(f"--fork-transaction-hash={transaction_hash}")

            self._setup_process(self._anvil_command)
            self._setup_w3()

        elif block_number is not None:
            # Otherwise, the fork can be reset in place without launching a new process
            fork_params = {}
            if block_number:
                fork_params["blockNumber"] = block_number

            method = "anvil_reset"
            resp = self.w3.provider.make_request(
                method=RPCEndpoint(method),
                params=[{"forking": fork_params}],
            )
            if "error" in resp:
                raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

        else:
            raise DegenbotValueError(message="No options provided.")

    def return_to_snapshot(self, snapshot_id: int) -> bool:
        if snapshot_id < 0:
            raise DegenbotValueError(message="ID cannot be negative")
        return bool(
            self.w3.provider.make_request(
                method=RPCEndpoint("evm_revert"),
                params=[snapshot_id],
            )["result"]
        )

    @validate_call
    def set_balance(
        self,
        address: str,
        balance: ValidatedUint256,
    ) -> None:
        self.w3.provider.make_request(
            method=RPCEndpoint("anvil_setBalance"),
            params=[address, hex(balance)],
        )

    def set_code(self, address: str, bytecode: bytes) -> None:
        self.w3.provider.make_request(
            method=RPCEndpoint("anvil_setCode"),
            params=[address, bytecode],
        )

    def set_coinbase(self, address: str) -> None:
        self.w3.provider.make_request(
            method=RPCEndpoint("anvil_setCoinbase"),
            params=[address],
        )

    def set_block_timestamp_interval(self, interval: int) -> None:
        self.w3.provider.make_request(
            method=RPCEndpoint("anvil_setBlockTimestampInterval"),
            params=[interval],
        )

    @validate_call
    async def set_next_base_fee_async(
        self,
        fee: ValidatedUint256,
    ) -> None:
        method = "anvil_setNextBlockBaseFeePerGas"
        async with self.async_w3 as async_w3:
            resp = await async_w3.provider.make_request(
                method=RPCEndpoint(method),
                params=[fee],
            )
            if "error" in resp:
                raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    @validate_call
    def set_next_base_fee(
        self,
        fee: ValidatedUint256,
    ) -> None:
        method = "anvil_setNextBlockBaseFeePerGas"
        resp = self.w3.provider.make_request(
            method=RPCEndpoint(method),
            params=[fee],
        )
        if "error" in resp:
            raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    @validate_call
    async def set_next_block_timestamp_async(
        self,
        timestamp: ValidatedUint256,
    ) -> None:
        method = "evm_setNextBlockTimestamp"
        async with self.async_w3 as async_w3:
            resp = await async_w3.provider.make_request(
                method=RPCEndpoint(method),
                params=[timestamp],
            )
            if "error" in resp:
                raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    @validate_call
    def set_next_block_timestamp(
        self,
        timestamp: ValidatedUint256,
    ) -> None:
        method = "evm_setNextBlockTimestamp"
        resp = self.w3.provider.make_request(
            method=RPCEndpoint(method),
            params=[timestamp],
        )
        if "error" in resp:
            raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    def set_nonce(self, address: str, nonce: int) -> None:
        method = "anvil_setNextBlockBaseFeePerGas"
        resp = self.w3.provider.make_request(
            method=RPCEndpoint(method),
            params=[address, nonce],
        )
        if "error" in resp:
            raise DegenbotError(message=f"RPC call to {method} returned error: {resp}")

    def set_snapshot(self) -> int:
        return int(
            self.w3.provider.make_request(
                method=RPCEndpoint("evm_snapshot"),
                params=[],
            )["result"],
            16,
        )

    def set_storage(
        self,
        address: HexAddress | bytes,
        position: int,
        value: HexStr | bytes | int,
    ) -> None:
        self.w3.provider.make_request(
            method=RPCEndpoint("anvil_setStorageAt"),
            params=[
                address,
                position,
                (
                    # Storage value must be padded to 32 bytes
                    HexBytes(value).hex().zfill(64)
                ),
            ],
        )



from dataclasses import dataclass

from eth_typing import ChecksumAddress

from degenbot.types.aliases import BlockNumber


@dataclass(slots=True, frozen=True, kw_only=True)
class AbstractPoolState:
    address: ChecksumAddress
    block: BlockNumber | None



from typing import TYPE_CHECKING, Any

from degenbot.exceptions import DegenbotError
from degenbot.types.aliases import BlockNumber

if TYPE_CHECKING:
    from degenbot.uniswap.v4_liquidity_pool import Hooks


class LiquidityPoolError(DegenbotError):
    """
    Exception raised inside liquidity pool helpers.
    """


# 2nd level exceptions for Liquidity Pool classes
class AddressMismatch(LiquidityPoolError):
    """
    The expected pool address does not match the provided address.
    """

    def __init__(self) -> None:
        super().__init__(message="Pool address verification failed.")


class LiquidityMapWordMissing(LiquidityPoolError):
    """
    A word bitmap is not included in the liquidity map.
    """

    def __init__(self, word: int) -> None:
        self.word = word
        super().__init__(message=f"Word {word} is unknown.")


class BrokenPool(LiquidityPoolError):
    def __init__(self) -> None:
        """
        Raised when an pool cannot or should not be built.
        """

        super().__init__(message="This pool is known to be broken.")


class ExternalUpdateError(LiquidityPoolError):
    """
    Raised when an external update does not pass sanity checks.
    """


class IncompleteSwap(LiquidityPoolError):
    """
    Raised if a swap calculation would not consume the input or deliver the requested output.
    """

    def __init__(self, amount_in: int, amount_out: int) -> None:
        self.amount_in = amount_in
        self.amount_out = amount_out
        super().__init__(message="Insufficient liquidity to swap for the requested amount.")

    def __reduce__(self) -> tuple[Any, ...]:
        # Pickling will raise an exception if a reduction method is not defined
        return self.__class__, (self.amount_in, self.amount_out)


class LateUpdateError(LiquidityPoolError):
    """
    Raised when an automatic update is attempted at a block prior to the last recorded update.
    """


class NoPoolStateAvailable(LiquidityPoolError):
    """
    Raised by the `restore_state_before_block` method when a previous pool state is not available.
    This can occur, e.g. if a pool was created in a block at or after a re-organization.
    """

    def __init__(self, block: BlockNumber) -> None:
        super().__init__(message=f"No pool state known prior to block {block}")


class InvalidSwapInputAmount(LiquidityPoolError):
    def __init__(self) -> None:
        """
        Raised if a swap input amount is invalid.
        """

        super().__init__(message="The swap input is invalid.")


class PossibleInaccurateResult(LiquidityPoolError):
    def __init__(self, amount_in: int, amount_out: int, hooks: set["Hooks"]) -> None:
        """
        Raised if a pool has an active hook that might invalidate the calculated result.
        """

        self.amount_in = amount_in
        self.amount_out = amount_out
        self.hooks = hooks
        super().__init__(
            message="The pool has one or more hooks that might invalidate the calculated result."
        )

    def __reduce__(self) -> tuple[Any, ...]:
        # Pickling will raise an exception if a reduction method is not defined
        return self.__class__, (self.amount_in, self.amount_out, self.hooks)



from typing import TYPE_CHECKING, Any

import eth_abi.abi
import eth_account.messages
import tenacity
from eth_typing import ChecksumAddress
from eth_utils.conversions import to_hex
from eth_utils.crypto import keccak
from hexbytes import HexBytes
from web3 import AsyncWeb3, Web3
from web3.eth.eth import Timeout
from web3.exceptions import Web3RPCError
from web3.types import BlockIdentifier, FilterParams, LogReceipt

from degenbot import get_checksum_address
from degenbot.constants import MAX_UINT256, MIN_UINT256
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.base import DegenbotError
from degenbot.exceptions.evm import InvalidUint256
from degenbot.logging import logger
from degenbot.types.aliases import BlockNumber

if TYPE_CHECKING:
    from eth_account.datastructures import SignedMessage


def create2_address(
    deployer: str | bytes,
    salt: bytes | str,
    init_code_hash: bytes | str,
) -> ChecksumAddress:
    """
    Generate the deterministic CREATE2 address for a given deployer, salt, and the keccak hash of
    the contract creation (init) bytecode.

    References:
        - https://eips.ethereum.org/EIPS/eip-1014
        - https://docs.openzeppelin.com/cli/2.8/deploying-with-create2
    """
    return get_checksum_address(
        keccak(HexBytes(0xFF) + HexBytes(deployer) + HexBytes(salt) + HexBytes(init_code_hash))[
            -20:
        ],  # Contract address is the least significant 20 bytes from the 32 byte hash
    )


def encode_function_calldata(
    function_prototype: str, function_arguments: list[Any] | None
) -> bytes:
    """
    Encode the calldata to execute a call to the given function prototype, with ordered arguments.
    The resulting bytes array will include the 4-byte function selector, followed by the
    ABI-encoded arguments.
    """

    if function_arguments is None:
        function_arguments = []

    return keccak(text=function_prototype)[:4] + eth_abi.abi.encode(
        types=extract_argument_types_from_function_prototype(function_prototype),
        args=function_arguments,
    )


def extract_argument_types_from_function_prototype(function_prototype: str) -> list[str]:
    """
    Extract the argument types from the function prototype.

    e.g. the argument types for the prototype 'function(address,uint256)' are ['address','uint256']
    """

    if function_args := function_prototype[
        function_prototype.find("(") + 1 : function_prototype.find(")") :
    ]:
        return function_args.split(",")

    return []


def eip_1167_clone_address(
    deployer: ChecksumAddress | str | bytes,
    implementation_contract: ChecksumAddress | str | bytes,
    salt: bytes,
) -> ChecksumAddress:
    """
    Calculate the contract address for an EIP-1167 minimal proxy contract deployed by `deployer`,
    using `salt`, delegating calls to the contract at `implementation` address.

    References:
        - https://github.com/ethereum/ercs/blob/master/ERCS/erc-1167.md
        - https://github.com/OpenZeppelin/openzeppelin-contracts/blob/master/contracts/proxy/Clones.sol
        - https://www.rareskills.io/post/eip-1167-minimal-proxy-standard-with-initialization-clone-pattern
    """

    minimal_proxy_code = (
        HexBytes("0x3d602d80600a3d3981f3")
        + HexBytes("0x363d3d373d3d3d363d73")
        + HexBytes(implementation_contract)
        + HexBytes("0x5af43d82803e903d91602b57fd5bf3")
    )

    return create2_address(
        deployer=deployer,
        salt=salt,
        init_code_hash=keccak(minimal_proxy_code),
    )


def eip_191_hash(message: str, private_key: str) -> str:
    """
    Get the signature hash (a hex-formatted string) for a given message and signing key.
    """
    result: SignedMessage = eth_account.Account.sign_message(
        signable_message=eth_account.messages.encode_defunct(
            text=to_hex(keccak(text=message)),
        ),
        private_key=private_key,
    )
    return result.signature.to_0x_hex()


def evm_divide(numerator: int, denominator: int) -> int:
    """
    Perform integer division, rounding towards zero to match the EVM behavior.
    """
    return -(-numerator // denominator) if numerator < 0 else numerator // denominator


def fetch_logs_retrying(
    w3: Web3,
    start_block: BlockNumber,
    end_block: BlockNumber,
    max_retries: int = 10,
    max_blocks_per_request: int | None = None,
    topic_signature: list[HexBytes] | None = None,
) -> list[LogReceipt]:
    """
    Fetch all event logs for the given topic signature (or all logs, if omitted), inclusive for the
    given block range.

    Max blocks per request is set to 5,000 if not specified.
    """

    if max_blocks_per_request is None:
        max_blocks_per_request = 5_000

    # The working block span is dynamic. It will be reduced quickly if timeouts occur, and increased
    # slowly following successful fetches
    working_span = max_blocks_per_request

    event_logs = []

    while True:
        try:
            for attempt in tenacity.Retrying(
                stop=tenacity.stop_after_attempt(max_retries),
                retry=tenacity.retry_if_exception_type(
                    (Timeout, Web3RPCError),
                ),
            ):
                chunk_end = min(end_block, start_block + working_span - 1)

                with attempt:
                    try:
                        logger.info(
                            f"Fetching logs for range {start_block}-{chunk_end} "
                            f" ({chunk_end - start_block + 1} blocks)"
                        )
                        event_logs.extend(
                            w3.eth.get_logs(
                                FilterParams(
                                    fromBlock=start_block,
                                    toBlock=chunk_end,
                                    topics=topic_signature,
                                )
                            )
                        )
                    except (Timeout, Web3RPCError):
                        # Decrease quickly (-50%)
                        working_span = max(
                            1,
                            working_span // 2,
                        )
                        logger.info(
                            f"Attempt {attempt.retry_state.attempt_number} timed out "
                            f"fetching {chunk_end - start_block + 1} blocks. "
                            f"Reducing to {working_span}..."
                        )
                        raise
                    else:
                        # Increase slowly up to the cap (+10%)
                        working_span = min(
                            working_span + working_span // 10,
                            max_blocks_per_request,
                        )

            if chunk_end == end_block:
                return event_logs

            start_block = chunk_end + 1

        except tenacity.RetryError:
            raise DegenbotError(
                message=f"Timed out fetching logs after {max_retries} tries."
            ) from None


def get_number_for_block_identifier(identifier: BlockIdentifier | None, w3: Web3) -> BlockNumber:
    match identifier:
        case None:
            return w3.eth.get_block_number()
        case int() as block_number_as_int:
            return block_number_as_int
        case "latest" | "earliest" | "pending" | "safe" | "finalized" as block_tag:
            block = w3.eth.get_block(block_tag)
            block_number = block.get("number")
            if TYPE_CHECKING:
                assert block_number is not None
            return block_number
        case str() as block_number_as_str:
            try:
                return int(block_number_as_str, 16)
            except ValueError:
                raise DegenbotValueError(
                    message=f"Invalid block identifier {identifier!r}"
                ) from None
        case bytes() as block_number_as_bytes:
            return int.from_bytes(block_number_as_bytes, byteorder="big")
        case _:
            raise DegenbotValueError(message=f"Invalid block identifier {identifier!r}")


async def get_number_for_block_identifier_async(
    identifier: BlockIdentifier | None, w3: AsyncWeb3
) -> BlockNumber:
    match identifier:
        case None:
            return await w3.eth.get_block_number()
        case int() as block_number_as_int:
            return block_number_as_int
        case "latest" | "earliest" | "pending" | "safe" | "finalized" as block_tag:
            block = await w3.eth.get_block(block_tag)
            block_number = block.get("number")
            if TYPE_CHECKING:
                assert block_number is not None
            return block_number
        case str() as block_number_as_str:
            try:
                return int(block_number_as_str, 16)
            except ValueError:
                raise DegenbotValueError(
                    message=f"Invalid block identifier {identifier!r}"
                ) from None
        case bytes() as block_number_as_bytes:
            return int.from_bytes(block_number_as_bytes, byteorder="big")
        case _:
            raise DegenbotValueError(message=f"Invalid block identifier {identifier!r}")


def next_base_fee(
    parent_base_fee: int,
    parent_gas_used: int,
    parent_gas_limit: int,
    min_base_fee: int | None = None,
    base_fee_max_change_denominator: int = 8,
    elasticity_multiplier: int = 2,
) -> int:
    """
    Calculate next base fee for an EIP-1559 compatible blockchain. The
    formula is taken from the example code in the EIP-1559 proposal (ref:
    https://eips.ethereum.org/EIPS/eip-1559).

    The default values for `base_fee_max_change_denominator` and
    `elasticity_multiplier` are taken from EIP-1559.

    Enforces `min_base_fee` if provided.
    """

    last_gas_target = parent_gas_limit // elasticity_multiplier

    if parent_gas_used == last_gas_target:
        _next_base_fee = parent_base_fee
    elif parent_gas_used > last_gas_target:
        gas_used_delta = parent_gas_used - last_gas_target
        base_fee_delta = max(
            parent_base_fee * gas_used_delta // last_gas_target // base_fee_max_change_denominator,
            1,
        )
        _next_base_fee = parent_base_fee + base_fee_delta
    else:
        gas_used_delta = last_gas_target - parent_gas_used
        base_fee_delta = (
            parent_base_fee * gas_used_delta // last_gas_target // base_fee_max_change_denominator
        )
        _next_base_fee = parent_base_fee - base_fee_delta

    return max(min_base_fee, _next_base_fee) if min_base_fee else _next_base_fee


def raise_if_invalid_uint256(number: int) -> None:
    if (MIN_UINT256 <= number <= MAX_UINT256) is False:
        raise InvalidUint256


def raw_call(
    w3: Web3,
    address: ChecksumAddress,
    calldata: bytes,
    return_types: list[str],
    block_identifier: BlockIdentifier | None = None,
) -> tuple[Any, ...]:
    """
    Perform an eth_call at the given address and returns the decoded response.
    """

    return eth_abi.abi.decode(
        types=return_types,
        data=w3.eth.call(
            transaction={
                "to": address,
                "data": calldata,
            },
            block_identifier=get_number_for_block_identifier(block_identifier, w3),
        ),
    )



import contextlib
from threading import Lock
from typing import TYPE_CHECKING, Any, Never, Self

from eth_typing import ChecksumAddress

from degenbot import connection_manager, get_checksum_address
from degenbot.aerodrome.functions import (
    generate_aerodrome_v2_pool_address,
    generate_aerodrome_v3_pool_address,
)
from degenbot.aerodrome.pools import AerodromeV2Pool, AerodromeV3Pool
from degenbot.exceptions.liquidity_pool import LiquidityPoolError
from degenbot.exceptions.manager import (
    ManagerAlreadyInitialized,
    PoolCreationFailed,
    PoolNotAssociated,
)
from degenbot.logging import logger
from degenbot.registry import pool_registry
from degenbot.types.abstract.pool_manager import AbstractPoolManager
from degenbot.types.aliases import ChainId
from degenbot.uniswap.deployments import FACTORY_DEPLOYMENTS, UniswapV2ExchangeDeployment
from degenbot.uniswap.managers import AbstractUniswapV3PoolManager
from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool


class _AbstractAerodromeV2PoolManager[Pool: AerodromeV2Pool](AbstractPoolManager[Pool]):
    """
    Abstract class used to inject the AerodromeV2Pool class into the parent abstract pool manager
    class where the tracking dicts are defined.
    """

    def get_pool(
        self,
        pool_address: ChecksumAddress | str,
        *,
        silent: bool = False,
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> Pool:
        """
        Get a pool from its address. If the pool is already tracked or found in the global registry,
        that instance will be returned. Otherwise, a new one will be built.
        """

        pool_address = get_checksum_address(pool_address)

        with contextlib.suppress(KeyError):
            return self._tracked_pools[pool_address]

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(pool_address)

        # Check if the pool registry already has this pool
        if (
            pool_from_registry := pool_registry.get(
                pool_address=pool_address,
                chain_id=self.chain_id,
            )
        ) is not None:
            if TYPE_CHECKING:
                assert isinstance(pool_from_registry, self.pool_factory)
            if pool_from_registry.factory == self._factory_address:
                self._add_tracked_pool(pool_from_registry)
                return pool_from_registry
            self._untracked_pools.add(pool_address)
            raise PoolNotAssociated(pool_address)

        try:
            new_pool = self.pool_factory(
                address=pool_address,
                silent=silent,
                **(pool_class_kwargs or {}),
            )
        except LiquidityPoolError as exc:
            raise PoolCreationFailed(
                message=f"Could not build V2 pool {pool_address}: {exc}"
            ) from exc
        else:
            self._add_tracked_pool(new_pool)
            return new_pool


class AerodromeV2PoolManager(
    _AbstractAerodromeV2PoolManager[AerodromeV2Pool], pool_factory=AerodromeV2Pool
):
    """
    A class that generates and tracks concrete instances of a Uniswap V2 liquidity pool helper or
    one of its child classes.
    """

    POOL_IMPLEMENTATION_ADDRESS = get_checksum_address("0xA4e46b4f701c62e14DF11B48dCe76A7d793CD6d7")

    @classmethod
    def from_exchange(
        cls,
        exchange: UniswapV2ExchangeDeployment,
    ) -> Self:
        return cls(
            factory_address=exchange.factory.address,
            deployer_address=exchange.factory.deployer,
            pool_init_hash=exchange.factory.pool_init_hash,
        )

    def __init__(
        self,
        *,
        factory_address: str,
        chain_id: ChainId | None = None,
        deployer_address: ChecksumAddress | str | None = None,
        pool_init_hash: str | None = None,
    ) -> None:
        factory_address = get_checksum_address(factory_address)

        if chain_id is None:
            chain_id = connection_manager.default_chain_id

        if (chain_id, factory_address) in self.instances:
            raise ManagerAlreadyInitialized(
                message="A manager has already been initialized for this address. Access it using the get_instance() class method"  # noqa:E501
            )
        self.instances[(chain_id, factory_address)] = self

        try:
            factory_deployment = FACTORY_DEPLOYMENTS[chain_id][factory_address]
            deployer_address = (
                factory_deployment.deployer
                if factory_deployment.deployer is not None
                else factory_address
            )
            pool_init_hash = factory_deployment.pool_init_hash
        except KeyError:
            if pool_init_hash is None:  # pragma: no branch
                logger.info("Pool init hash is unknown. Using Uniswap V3 mainnet default.")
                pool_init_hash = UniswapV2Pool.UNISWAP_V2_MAINNET_POOL_INIT_HASH
            deployer_address = (
                get_checksum_address(deployer_address)
                if deployer_address is not None
                else factory_address
            )

        self._lock = Lock()
        self._chain_id = chain_id
        self._deployer_address = deployer_address
        self._factory_address = factory_address
        self._pool_init_hash = pool_init_hash
        self._tracked_pools = {}
        self._untracked_pools: set[ChecksumAddress] = set()

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(factory={self._factory_address})"

    def get_stable_pool(
        self,
        token_addresses: tuple[str, str],
        *,
        silent: bool = False,
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> AerodromeV2Pool:
        """
        Get a stable pool by its token addresses. The token addresses may be passed in any order.
        """

        pool_address = generate_aerodrome_v2_pool_address(
            deployer_address=self._deployer_address,
            token_addresses=sorted(token_addresses),
            implementation_address=self.POOL_IMPLEMENTATION_ADDRESS,
            stable=True,
        )

        return self.get_pool(
            pool_address=pool_address,
            silent=silent,
            pool_class_kwargs=pool_class_kwargs,
        )

    def get_volatile_pool(
        self,
        token_addresses: tuple[str, str],
        *,
        silent: bool = False,
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> AerodromeV2Pool:
        """
        Get a volatile pool by its token addresses. The token addresses may be passed in any order.
        """

        pool_address = generate_aerodrome_v2_pool_address(
            deployer_address=self._deployer_address,
            token_addresses=sorted(token_addresses),
            implementation_address=self.POOL_IMPLEMENTATION_ADDRESS,
            stable=False,
        )

        return self.get_pool(
            pool_address=pool_address,
            silent=silent,
            pool_class_kwargs=pool_class_kwargs,
        )


class AerodromeV3PoolManager(
    AbstractUniswapV3PoolManager[AerodromeV3Pool], pool_factory=AerodromeV3Pool
):
    POOL_IMPLEMENTATION_ADDRESS = get_checksum_address("0xeC8E5342B19977B4eF8892e02D8DAEcfa1315831")

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(factory={self._factory_address})"

    def get_pool_from_tokens_and_fee(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Never:
        raise NotImplementedError

    def get_pool_from_tokens_and_tick_spacing(
        self,
        token_addresses: tuple[str, str],
        tick_spacing: int,
        *,
        silent: bool = False,
        # keyword arguments passed to the pool class constructor
        pool_class_kwargs: dict[str, Any] | None = None,
    ) -> AerodromeV3Pool:
        pool_address = generate_aerodrome_v3_pool_address(
            deployer_address=self._deployer_address,
            token_addresses=sorted(token_addresses),
            implementation_address=self.POOL_IMPLEMENTATION_ADDRESS,
            tick_spacing=tick_spacing,
        )

        pool = self.get_pool(
            pool_address=pool_address,
            silent=silent,
            pool_class_kwargs=pool_class_kwargs,
        )
        assert isinstance(pool, AerodromeV3Pool)
        return pool



from degenbot.constants import MAX_UINT256, MIN_UINT256
from degenbot.exceptions.evm import EVMRevertError

# This module is adapted from the Uniswap V3 BitMath.sol library.
# Reference: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/BitMath.sol


def least_significant_bit(number: int) -> int:
    """
    Find the least significant bit for the given number.

    This function is rewritten to use simple string manipulation instead of the binary search
    implemented by the official Solidity contract.
    """

    if number <= MIN_UINT256:
        raise EVMRevertError(error="required: number > 0")
    if number > MAX_UINT256:
        raise EVMRevertError(error="required: number <= max(uint256)")

    """
    Reverse the binary string and search for LSB by returning the position of the first "1" value

    e.g. bin(69) == '0b1000101', trim the '0b' and reverse the string
    '1010001'
           ^---- LSB @ position 0
    """

    num_string = bin(number)[2:][::-1]
    return num_string.find("1")


def most_significant_bit(number: int) -> int:
    """
    Find the most significant bit for the given number.

    This function is rewritten to use simple string manipulation instead of the binary search
    implemented by the official Solidity contract.
    """

    if number <= MIN_UINT256:
        raise EVMRevertError(error="required: number >= 0 ")
    if number > MAX_UINT256:
        raise EVMRevertError(error="required: number <= max(uint256) ")

    """
    Search for MSB by returning the position of the first "1" value

    e.g. bin(69) == '0b1000101', trim the '0b'
    '1000101'
     ^-------- MSB @ index 0 (position 7 after reversing)
    """

    num_string = bin(number)[2:]
    return (len(num_string) - 1) - num_string.find("1")
